# Copyright © 2025 UChicago Argonne, LLC All right reserved
# Full license accessible at https://github.com//AdvancedPhotonSource/pty-chi/blob/main/LICENSE

from typing import Union
from ptychi.movies.settings import (
    ObjectMovieSettings,
    ProbeMovieSettings,
    ProbePlotTypes,
    ProcessFunctionType,
)
import ptychi.reconstructors.base as base
import torch
import numpy as np

movie_setting_types = Union[ObjectMovieSettings, ProbeMovieSettings]


def prepare_movie_subject(
    reconstructor: "base.Reconstructor",
    movie_settings: movie_setting_types,
) -> np.ndarray:
    if isinstance(movie_settings, ObjectMovieSettings):
        array_out = prepare_object(reconstructor, movie_settings)
    elif isinstance(movie_settings, ProbeMovieSettings):
        array_out = prepare_probe(reconstructor, movie_settings)

    scale = movie_settings.snapshot.scale
    array_out = array_out.cpu().detach().numpy()[::scale, ::scale]
    return array_out


def process_function(array: torch.Tensor, process_function_type: ProcessFunctionType):
    if process_function_type is ProcessFunctionType.MAGNITUDE:
        return array.abs()
    elif process_function_type is ProcessFunctionType.PHASE:
        return array.angle()


def prepare_object(reconstructor: "base.Reconstructor", movie_settings: ObjectMovieSettings):
    array_out = reconstructor.parameter_group.object.data[movie_settings.slice_index]
    array_out = process_function(array_out, movie_settings.process_function)
    return array_out


def prepare_probe(reconstructor: "base.Reconstructor", movie_settings: ProbeMovieSettings):
    array_out = reconstructor.parameter_group.probe.data
    if movie_settings.plot_type is ProbePlotTypes.INCOHERENT_SUM:
        array_out = (array_out[0].abs() ** 2).sum(0)
    elif movie_settings.plot_type is ProbePlotTypes.SEPERATE_MODES:
        if movie_settings.mode_indices is None:
            mode_idx = range(0, reconstructor.parameter_group.probe.n_modes)
        else:
            mode_idx = movie_settings.mode_indices
        array_out = array_out[0, mode_idx]
        array_out = process_function(array_out, movie_settings.process_function)
        # Reshape so each probe mode is next to eachother
        probe_width = array_out.shape[1]
        array_out = array_out.permute((0, 2, 1))
        array_out = array_out.reshape(shape=(-1, probe_width))
        array_out = array_out.transpose(1, 0)

    return array_out



# Copyright © 2025 UChicago Argonne, LLC All right reserved
# Full license accessible at https://github.com//AdvancedPhotonSource/pty-chi/blob/main/LICENSE

from .ad_general import AutodiffReconstructor
from .ad_ptychography import AutodiffPtychographyReconstructor
from .lsqml import LSQMLReconstructor
from .pie import PIEReconstructor, EPIEReconstructor, RPIEReconstructor
from .dm import DMReconstructor
from .bh import BHReconstructor

__all__ = [
    "AutodiffReconstructor",
    "AutodiffPtychographyReconstructor",
    "LSQMLReconstructor",
    "PIEReconstructor",
    "EPIEReconstructor",
    "RPIEReconstructor",
    "DMReconstructor",
    "BHReconstructor"
]



# Copyright © 2025 UChicago Argonne, LLC All right reserved
# Full license accessible at https://github.com//AdvancedPhotonSource/pty-chi/blob/main/LICENSE

from . import components
from . import models
from .models import unet, autoencoder


__all__ = ["components", "models", "unet", "autoencoder"]



import argparse

import torch
import numpy as np

from ptychi.data_structures.object import PlanarObject
from ptychi.api.options.base import ObjectOptions, RemoveGridArtifactsOptions
import ptychi.api as api

import test_utils as tutils


class TestRemoveGridArtifacts(tutils.BaseTester):

    def test_remove_grid_artifacts(self):
        phase = torch.zeros([64, 64])
        phase[::10, ::10] = 1
        data = torch.ones([1, 64, 64]) * torch.exp(1j * phase)
        object = PlanarObject(
            data=data,
            options=ObjectOptions(
                pixel_size_m=1,
                remove_grid_artifacts=RemoveGridArtifactsOptions(
                    enabled=True,
                    period_x_m=10,
                    period_y_m=10,
                    window_size=3,
                    direction=api.Directions.XY,
                ),
            ),
        )
        with torch.no_grad():
            object.remove_grid_artifacts()
        
        data = object.data.detach().cpu().numpy()[0]

        if self.debug:
            import matplotlib.pyplot as plt

            _, ax = plt.subplots()
            ax.imshow(np.angle(data))
            plt.show()
        
        assert np.max(np.abs((np.angle(data)))) < 0.1


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('--generate-gold', action='store_true')
    args = parser.parse_args()
    
    tester = TestRemoveGridArtifacts()
    tester.setup_method(name="", generate_data=False, generate_gold=args.generate_gold, debug=True)
    tester.test_remove_grid_artifacts()



import argparse

import torch
import numpy as np

import ptychi.api as api
from ptychi.api.task import PtychographyTask
from ptychi.utils import get_suggested_object_size, get_default_complex_dtype

import test_utils as tutils


class TestMultislicePtychoLSQMLNonsqPsize(tutils.BaseTester):

    @tutils.BaseTester.wrap_recon_tester(name='test_multislice_ptycho_lsqml_nonsq_psize')
    def test_multislice_ptycho_lsqml_nonsq_psize(self):
        self.setup_ptychi(cpu_only=False)
        
        data, probe, pixel_size_m, positions_px = self.load_data_ptychodus(
            *self.get_default_input_data_file_paths('multislice_ptycho_AuNi_nonsq_psize'),
            subtract_position_mean=True
        )
        pixel_width_m = pixel_size_m
        positions_px[:, 0] = positions_px[:, 0] / 2
        
        wavelength_m = 1.03e-10
        
        options = api.LSQMLOptions()
        
        options.data_options.data = data
        options.data_options.wavelength_m = wavelength_m
        
        options.object_options.initial_guess = torch.ones([2, *get_suggested_object_size(positions_px, probe.shape[-2:], extra=50)], dtype=get_default_complex_dtype())
        options.object_options.pixel_size_m = pixel_width_m
        options.object_options.pixel_size_aspect_ratio = 0.5
        options.object_options.slice_spacings_m = np.array([1e-5])
        options.object_options.optimizable = True
        options.object_options.optimizer = api.Optimizers.SGD
        options.object_options.step_size = 1
        
        options.probe_options.initial_guess = probe
        options.probe_options.optimizable = True
        options.probe_options.optimizer = api.Optimizers.SGD
        options.probe_options.step_size = 1
        
        options.probe_position_options.position_x_px = positions_px[:, 1]
        options.probe_position_options.position_y_px = positions_px[:, 0]
        options.probe_position_options.optimizable = False
        options.probe_position_options.optimizer = api.Optimizers.SGD
        options.probe_position_options.step_size = 1e-1
        options.probe_position_options.correction_options.update_magnitude_limit = 5.0
        
        options.reconstructor_options.displayed_loss_function = api.LossFunctions.MSE_SQRT
        options.reconstructor_options.batch_size = 101
        options.reconstructor_options.num_epochs = 32
        options.reconstructor_options.default_device = api.Devices.GPU
        options.reconstructor_options.random_seed = 123
        options.reconstructor_options.allow_nondeterministic_algorithms = False
        
        task = PtychographyTask(options)
        task.run()

        recon = task.get_data_to_cpu('object', as_numpy=True)
        return recon
        
    
if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--generate-gold', action='store_true')
    args = parser.parse_args()

    tester = TestMultislicePtychoLSQMLNonsqPsize()
    tester.setup_method(name="", generate_data=False, generate_gold=args.generate_gold, debug=True)
    tester.test_multislice_ptycho_lsqml_nonsq_psize()



# Copyright © 2025 UChicago Argonne, LLC All right reserved
# Full license accessible at https://github.com//AdvancedPhotonSource/pty-chi/blob/main/LICENSE

from typing import Optional, Union, Tuple, Sequence, TYPE_CHECKING
import logging

import torch
from torch import Tensor
from torch.nn import Module, Parameter
from numpy import ndarray

from ptychi.utils import to_tensor
import ptychi.maps as maps
import ptychi.api.options.base as obase

if TYPE_CHECKING:
    import ptychi.api as api

logger = logging.getLogger(__name__)


class ComplexTensor(Module):
    """
    A module that stores the real and imaginary parts of a complex tensor
    as real tensors.

    The support of PyTorch DataParallel on complex parameters is flawed. To
    avoid the issue, complex parameters are stored as two real tensors.
    """

    def __init__(
        self, 
        data: Union[Tensor, ndarray], 
        requires_grad: bool = True, 
        data_as_parameter: bool = True, 
        *args, **kwargs
    ) -> None:
        super().__init__(*args, **kwargs)
        data = to_tensor(data)
        data = torch.stack([data.real, data.imag], dim=-1).requires_grad_(requires_grad)
        data = data.type(torch.get_default_dtype())

        if data_as_parameter:
            self.register_parameter(name="data", param=Parameter(data))
        else:
            self.register_buffer("data", data)

    def mag(self) -> Tensor:
        return torch.sqrt(self.data[..., 0] ** 2 + self.data[..., 1] ** 2)

    def magsq(self) -> Tensor:
        return self.data[..., 0] ** 2 + self.data[..., 1] ** 2

    def phase(self) -> Tensor:
        return torch.atan2(self.data[..., 1], self.data[..., 0])

    def real(self) -> Tensor:
        return self.data[..., 0]

    def imag(self) -> Tensor:
        return self.data[..., 1]

    def complex(self) -> Tensor:
        return self.real() + 1j * self.imag()

    @property
    def shape(self) -> Tuple[int, ...]:
        return self.data.shape[:-1]

    def set_data(self, data: Union[Tensor, ndarray], slicer=None):
        if slicer is None:
            slicer = (slice(None),)
        elif not isinstance(slicer, Sequence):
            slicer = (slicer,)
        data = to_tensor(data)
        data = torch.stack([data.real, data.imag], dim=-1)
        data = data.type(torch.get_default_dtype())
        self.data[*slicer].copy_(to_tensor(data))


class ReconstructParameter(Module):
    name = None
    optimizable: bool = True
    optimization_plan: "api.OptimizationPlan" = None
    optimizer = None
    is_dummy = False

    def __init__(
        self,
        shape: Optional[Tuple[int, ...]] = None,
        data: Optional[Union[Tensor, ndarray]] = None,
        is_complex: bool = False,
        name: Optional[str] = None,
        options: "api.options.base.ParameterOptions" = None,
        data_as_parameter: bool = True,
        build_optimizer: bool = True,
        *args,
        **kwargs,
    ) -> None:
        """The base reconstructor parameter class.

        Parameters
        ----------
        shape : Optional[Tuple[int, ...]], optional
            The shape of the parameter.
        data : Optional[Union[Tensor, ndarray]], optional
            The data of the parameter.
        is_complex : bool, optional
            Whether the parameter is complex.
        name : Optional[str], optional
            The name of the parameter.
        options : api.options.base.ParameterOptions, optional
            Options of the parameter.
        data_as_parameter : bool, optional
            Whether the data is stored as a torch.Parameter. In most cases this should be True,
            but for DIPObject, the data is not directly optimized so it should just be a buffer.
        """
        super().__init__(*args, **kwargs)
        if shape is None and data is None:
            raise ValueError("Either shape or data must be specified.")
        if options is None:
            if isinstance(self, DummyParameter):
                options = obase.ParameterOptions(optimizable=False)
            else:
                raise ValueError("Parameter options of {} must be specified.".format(self.name))

        self.name = name
        self.options = options
        self.optimizable = self.options.optimizable
        self.optimization_plan = self.options.optimization_plan
        if self.optimization_plan is None:
            raise ValueError("Optimization plan of {} is not specified.".format(self.name))
        self.optimizer_class = maps.get_optimizer_by_enum(self.options.optimizer)

        self.optimizer_params = (
            {} if self.options.optimizer_params is None else self.options.optimizer_params
        )
        # If optimizer_params has 'lr', it will overwrite the step_size.
        self.optimizer_params = dict(
            {"lr": self.options.step_size}, **self.options.optimizer_params
        )
        self.optimizer = None
        
        self.sub_modules = []
        self.optimizable_sub_modules = []
        self.is_complex = is_complex
        self.preconditioner = None
        self.update_buffer = None

        if is_complex:
            if data is not None:
                self.tensor = ComplexTensor(data, data_as_parameter=data_as_parameter).requires_grad_(self.optimizable)
            else:
                self.tensor = ComplexTensor(torch.zeros(shape), data_as_parameter=data_as_parameter, requires_grad=self.optimizable)
        else:
            if data is not None:
                tensor = to_tensor(data).requires_grad_(self.optimizable)
            else:
                tensor = torch.zeros(shape).requires_grad_(self.optimizable)
            # Register the tensor as a parameter. In subclasses, do the same for any
            # additional differentiable parameters. If you have a buffer that does not
            # need gradients, use register_buffer instead.
            if data_as_parameter:
                self.register_parameter("tensor", Parameter(tensor))
            else:
                self.register_buffer("tensor", tensor)

        if build_optimizer:
            self.build_optimizer()

    @property
    def shape(self) -> Tuple[int, ...]:
        return self.tensor.shape

    @property
    def data(self) -> Tensor:
        """Get a copy of the parameter data. For complex parameters,
        it returns a complex tensor rather than real tensors giving the
        real and imaginary parts as in the internal representation
        of ComplexTensor.

        Returns
        -------
        Tensor
            A copy of the parameter data.
        """
        if self.is_complex:
            return self.tensor.complex()
        else:
            return self.tensor.clone()
        
    def register_optimizable_sub_module(self, sub_module):
        if sub_module.optimizable and sub_module not in self.optimizable_sub_modules:
            self.optimizable_sub_modules.append(sub_module)
        if sub_module not in self.sub_modules:
            self.sub_modules.append(sub_module)

    def build_optimizer(self):
        if self.optimizable and self.optimizer_class is None:
            raise ValueError(
                "Parameter {} is optimizable but no optimizer is specified.".format(self.name)
            )
        if self.optimizable:
            if isinstance(self.tensor, ComplexTensor):
                self.optimizer = self.optimizer_class([self.tensor.data], **self.optimizer_params)
            else:
                self.optimizer = self.optimizer_class([self.tensor], **self.optimizer_params)

    def set_optimizable(self, optimizable):
        self.optimizable = optimizable
        self.tensor.requires_grad_(optimizable)

    def get_tensor(self, name):
        """Get a member tensor in this object.

        It is necessary to use this method to access memebers when
        # (1) the forward model is wrapped in DataParallel,
        # (2) multiple deivces are used,
        # (3) the model has complex parameters.
        # DataParallel adds an additional dimension at the end of each registered
        # complex parameter (not an issue for real parameters).
        This method selects the right index along that dimension by checking
        the device ID.
        """
        var = getattr(self, name)
        # If the current shape has one more dimension than the original shape,
        # it means that the DataParallel wrapper has added an additional
        # dimension. Select the right index from the last dimension.
        if len(var.shape) > len(self.shape):
            dev_id = var.device.index
            if dev_id is None:
                raise RuntimeError("Expecting multi-GPU, but unable to find device ID.")
            var = var[..., dev_id]
        return var

    def set_data(
        self, data, slicer: Optional[Union[slice, int] | tuple[Union[slice, int], ...]] = None
    ):
        if slicer is None:
            slicer = (slice(None),)
        elif not isinstance(slicer, Sequence):
            slicer = (slicer,)
        if isinstance(self.tensor, ComplexTensor):
            self.tensor.set_data(data, slicer=slicer)
        else:
            self.tensor[*slicer].copy_(to_tensor(data))

    def get_grad(self):
        if isinstance(self.tensor, ComplexTensor):
            return self.tensor.data.grad[..., 0] + 1j * self.tensor.data.grad[..., 1]
        else:
            return self.tensor.grad

    def set_grad(
        self,
        grad: Tensor,
        slicer: Optional[Union[slice, int] | tuple[Union[slice, int], ...]] = None,
    ):
        """
        Populate the `grad` field of the contained tensor, so that it can optimized
        by PyTorch optimizers. You should not need this for AutodiffReconstructor.
        However, method without automatic differentiation needs this to fill in the gradients
        manually.

        Parameters
        ----------
        grad : Tensor
            A tensor giving the gradient. If the gradient is complex, give it as it is.
            This routine will separate the real and imaginary parts and write them into
            the tensor.grad inside the ComplexTensor object.
        slicer : Optional[Union[slice, int] | tuple[Union[slice, int], ...]]
            A tuple of, or a single slice object or integer, that defines the region of
            the region of the gradient to update. The shape of `grad` should match
            the region given by `slicer`, if given. If None, the whole gradient is updated.
        """
        if self.tensor.data.grad is None and slicer is not None:
            raise ValueError("Setting gradient with slicing is not allowed when gradient is None.")
        if slicer is None:
            slicer = (slice(None),)
        elif not isinstance(slicer, Sequence):
            slicer = (slicer,)
        if len(slicer) > len(self.shape):
            raise ValueError("The number of slices should not exceed the number of dimensions.")
        if isinstance(self.tensor, ComplexTensor):
            grad = torch.stack([grad.real, grad.imag], dim=-1)
            if self.tensor.data.grad is None:
                self.tensor.data.grad = grad
            else:
                self.tensor.data.grad[*slicer, ..., :] = grad
        else:
            if self.tensor.grad is None:
                self.tensor.grad = grad
            else:
                self.tensor.grad[*slicer] = grad

    def initialize_grad(self):
        """
        Initialize the gradient with zeros.
        """
        if isinstance(self.tensor, ComplexTensor):
            self.tensor.data.grad = torch.zeros_like(self.tensor.data)
        else:
            self.tensor.grad = torch.zeros_like(self.tensor)

    def post_update_hook(self, *args, **kwargs):
        pass

    def optimization_enabled(self, epoch: int):
        if self.optimizable and self.optimization_plan.is_enabled(epoch):
            enabled = True
        else:
            enabled = False
        logger.debug(f"{self.name} optimization enabled at epoch {epoch}: {enabled}")
        return enabled

    def get_config_dict(self):
        return self.options.get_non_data_fields()
    
    def step_optimizer(self, limit: float = None):
        """Step the optimizer with gradient filled in. This function
        can optionally impose a limit on the magnitude of the update.

        Parameters
        ----------
        limit : float, optional
            The maximum allowed magnitude of the update. Set to None to disable the limit.
        """
        if limit is not None and limit <= 0:
            raise ValueError("`limit` should either be None or a positive number.")
        if limit == torch.inf:
            limit = None
        if limit is not None:
            data0 = self.data
        self.optimizer.step()
        if limit is not None:
            data = self.data
            dx = data - data0
            update_mag = dx.abs()
            exceed_mask = update_mag > limit
            dx[exceed_mask] = dx[exceed_mask] * limit / update_mag[exceed_mask]
            self.set_data(data0 + dx)


class DummyParameter(ReconstructParameter):
    is_dummy = True

    def __init__(self, *args, **kwargs):
        super().__init__(shape=(1,), *args, **kwargs)

    def optimization_enabled(self, *args, **kwargs):
        return False


class BoundingBox(torch.nn.Module):
    def __init__(self, sy, ey, sx, ex, origin=(0, 0)):
        super().__init__()
        tensor = to_tensor([sy, ey, sx, ex])
        origin = to_tensor(origin)
        self.register_buffer("tensor", tensor)
        self.register_buffer("origin", origin)

    def __repr__(self):
        return __class__.__name__ + "(sy={}, ey={}, sx={}, ex={}, origin={})".format(
            self.sy, self.ey, self.sx, self.ex, self.origin
        )

    @property
    def sy(self):
        return self.tensor[0]

    @property
    def ey(self):
        return self.tensor[1]

    @property
    def sx(self):
        return self.tensor[2]

    @property
    def ex(self):
        return self.tensor[3]

    def get_bbox_with_top_left_origin(self) -> "BoundingBox":
        """
        Get a new bounding box with the top left origin.
        """
        bbox = BoundingBox(*(self.origin.repeat_interleave(2) + self.tensor), origin=(0, 0))
        return bbox

    def get_slicer(self) -> slice:
        return (slice(int(self.sy), int(self.ey)), slice(int(self.sx), int(self.ex)))



import argparse
import os

import torch
import numpy as np

import ptychi.api as api
from ptychi.api.task import PtychographyTask
from ptychi.utils import get_suggested_object_size, get_default_complex_dtype

import test_utils as tutils


class Test2DPtychoRPIE_SDL(tutils.TungstenDataTester):
    @tutils.TungstenDataTester.wrap_recon_tester(name="test_2d_ptycho_rpie_synthesisdictlearn")
    def test_2d_ptycho_rpie_synthesisdictlearn(self):
        self.setup_ptychi(cpu_only=False)

        data, probe, pixel_size_m, positions_px = self.load_tungsten_data(additional_opr_modes=0)

        npz_dict_file = np.load(
            os.path.join(
                self.get_ci_input_data_dir(), "zernike2D_dictionaries", "testing_sdl_dictionary.npz"
            )
        )
        D = npz_dict_file["a"]
        npz_dict_file.close()

        options = api.RPIEOptions()
        options.data_options.data = data

        options.object_options.initial_guess = torch.ones(
            [1, *get_suggested_object_size(positions_px, probe.shape[-2:], extra=100)],
            dtype=get_default_complex_dtype(),
        )
        options.object_options.pixel_size_m = pixel_size_m
        options.object_options.optimizable = True
        options.object_options.optimizer = api.Optimizers.SGD
        options.object_options.step_size = 1e-1
        options.object_options.alpha = 1e-0

        options.probe_options.initial_guess = probe
        options.probe_options.optimizable = True
        options.probe_options.optimizer = api.Optimizers.SGD
        options.probe_options.orthogonalize_incoherent_modes.enabled = True
        options.probe_options.step_size = 1e-0
        options.probe_options.alpha = 1e-0

        options.probe_options.experimental.sdl_probe_options.enabled = True
        options.probe_options.experimental.sdl_probe_options.d_mat = np.asarray(
            D, dtype=np.complex64
        )
        options.probe_options.experimental.sdl_probe_options.d_mat_conj_transpose = np.conj(
            options.probe_options.experimental.sdl_probe_options.d_mat
        ).T
        options.probe_options.experimental.sdl_probe_options.d_mat_pinv = np.linalg.pinv(
            options.probe_options.experimental.sdl_probe_options.d_mat
        )
        options.probe_options.experimental.sdl_probe_options.probe_sparse_code_nnz = np.round(
            0.90 * D.shape[-1]
        )

        options.probe_position_options.position_x_px = positions_px[:, 1]
        options.probe_position_options.position_y_px = positions_px[:, 0]
        options.probe_position_options.optimizable = False

        options.reconstructor_options.batch_size = round(data.shape[0] * 0.1)
        options.reconstructor_options.num_epochs = 50
        options.reconstructor_options.allow_nondeterministic_algorithms = False

        task = PtychographyTask(options)
        task.run()

        recon = task.get_data_to_cpu("object", as_numpy=True)[0]

        return recon


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--generate-gold", action="store_true")
    args = parser.parse_args()

    tester = Test2DPtychoRPIE_SDL()
    tester.setup_method(name="", generate_data=False, generate_gold=args.generate_gold, debug=True)
    tester.test_2d_ptycho_rpie_synthesisdictlearn()



# Copyright © 2025 UChicago Argonne, LLC All right reserved
# Full license accessible at https://github.com//AdvancedPhotonSource/pty-chi/blob/main/LICENSE

from typing import Optional, Type, Union
import dataclasses
from dataclasses import field


import ptychi.api.options.base as base
import ptychi.api.options.task as task_options
import ptychi.api.enums as enums
import ptychi.forward_models as fm


@dataclasses.dataclass
class AutodiffReconstructorOptions(base.ReconstructorOptions):
    loss_function: enums.LossFunctions = enums.LossFunctions.MSE_SQRT
    """
    The loss function.
    """

    forward_model_class: Union[enums.ForwardModels, Type["fm.ForwardModel"]] = enums.ForwardModels.base
    """
    The forward model class.
    """

    forward_model_params: Optional[dict] = None
    """
    The forward model parameters.
    """

    def get_reconstructor_type(self) -> enums.Reconstructors:
        return enums.Reconstructors.AD_GENERAL


@dataclasses.dataclass
class AutodiffOptions(task_options.PtychographyTaskOptions):
    reconstructor_options: AutodiffReconstructorOptions = field(
        default_factory=AutodiffReconstructorOptions
    )



import argparse

import torch
import numpy as np

import ptychi.api as api
from ptychi.api.task import PtychographyTask
from ptychi.utils import get_suggested_object_size, get_default_complex_dtype

import test_utils as tutils


class Test2DPtychoDM(tutils.TungstenDataTester):
    @tutils.TungstenDataTester.wrap_recon_tester(name="test_2d_ptycho_dm")
    def test_2d_ptycho_dm(self):
        self.setup_ptychi(cpu_only=False)

        data, probe, pixel_size_m, positions_px = self.load_tungsten_data(additional_opr_modes=0)

        options = api.DMOptions()
        options.data_options.data = data

        options.object_options.initial_guess = torch.ones(
            [1, *get_suggested_object_size(positions_px, probe.shape[-2:], extra=100)],
            dtype=get_default_complex_dtype(),
        )
        options.object_options.pixel_size_m = pixel_size_m
        options.object_options.optimizable = True

        options.probe_options.initial_guess = probe
        options.probe_options.power_constraint.probe_power = np.sum(
            np.max(data, axis=-3), axis=(-2, -1)
        )
        options.probe_options.power_constraint.enabled = True
        options.probe_options.optimizable = True

        options.probe_position_options.position_x_px = positions_px[:, 1]
        options.probe_position_options.position_y_px = positions_px[:, 0]
        options.probe_position_options.optimizable = False

        options.reconstructor_options.num_epochs = 8
        options.reconstructor_options.chunk_length = 100
        options.reconstructor_options.allow_nondeterministic_algorithms = False
        task = PtychographyTask(options)
        task.run()

        recon = task.get_data_to_cpu("object", as_numpy=True)[0]
        return recon


class Test2DPtychoDMPositionCorrection(tutils.TungstenDataTester):
    @tutils.TungstenDataTester.wrap_recon_tester(name="test_2d_ptycho_dm_position_correction")
    def test_2d_ptycho_dm_position_correction(self):
        self.setup_ptychi(cpu_only=False)

        data, probe, pixel_size_m, positions_px = self.load_tungsten_data(additional_opr_modes=0)

        options = api.DMOptions()
        options.data_options.data = data

        options.object_options.initial_guess = torch.ones(
            [1, *get_suggested_object_size(positions_px, probe.shape[-2:], extra=100)],
            dtype=get_default_complex_dtype(),
        )
        options.object_options.pixel_size_m = pixel_size_m
        options.object_options.optimizable = True

        options.probe_options.initial_guess = probe
        options.probe_options.power_constraint.probe_power = np.sum(
            np.max(data, axis=-3), axis=(-2, -1)
        )
        options.probe_options.power_constraint.enabled = True
        options.probe_options.optimizable = True

        options.probe_position_options.position_x_px = positions_px[:, 1]
        options.probe_position_options.position_y_px = positions_px[:, 0]
        options.probe_position_options.optimizable = True

        options.reconstructor_options.num_epochs = 8
        options.reconstructor_options.chunk_length = 100
        options.reconstructor_options.allow_nondeterministic_algorithms = False
        task = PtychographyTask(options)
        task.run()

        recon = task.get_data_to_cpu("object", as_numpy=True)[0]
        return recon


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--generate-gold", action="store_true")
    args = parser.parse_args()

    tester = Test2DPtychoDM()
    tester.setup_method(name="", generate_data=False, generate_gold=args.generate_gold, debug=True)
    tester.test_2d_ptycho_dm()

    tester=Test2DPtychoDMPositionCorrection()
    tester.setup_method(name="", generate_data=False, generate_gold=args.generate_gold, debug=True)
    tester.test_2d_ptycho_dm_position_correction()



import argparse
import logging

import torch
import numpy as np

import ptychi.api as api
from ptychi.api.task import PtychographyTask
from ptychi.api import LSQMLOptions
import ptychi.utils as utils

import test_utils as tutils


class Test2dPtychoOptPlan(tutils.TungstenDataTester):

    @tutils.TungstenDataTester.wrap_recon_tester(name='test_2d_ptycho_opt_plan')
    def test_2d_ptycho_opt_plan(self):
        self.setup_ptychi(cpu_only=False)

        data, probe, pixel_size_m, positions_px = self.load_tungsten_data(pos_type='true', additional_opr_modes=3)
        object_init = torch.ones(
            [1, *utils.get_suggested_object_size(positions_px, probe.shape[-2:], extra=100)], 
            dtype=utils.get_default_complex_dtype()
        )
        
        options = LSQMLOptions()
        options.data_options.data = data
        
        options.object_options.initial_guess = object_init
        options.object_options.pixel_size_m = pixel_size_m
        options.object_options.optimizable = True
        options.object_options.optimizer = api.Optimizers.SGD
        options.object_options.step_size = 1
        
        options.probe_options.initial_guess = probe
        options.probe_options.optimizable = True
        options.probe_options.optimization_plan.start = 2
        options.probe_options.optimization_plan.end = None
        options.probe_options.optimizer = api.Optimizers.SGD
        options.probe_options.step_size = 1
        
        options.probe_position_options.position_x_px = positions_px[:, 1]
        options.probe_position_options.position_y_px = positions_px[:, 0]
        options.probe_position_options.correction_options.update_magnitude_limit = 1.0
        options.probe_position_options.optimizable = True
        options.probe_position_options.optimization_plan.start = 1
        options.probe_position_options.optimization_plan.end = 3
        options.probe_position_options.optimization_plan.stride = 2
        options.probe_position_options.optimizer = api.Optimizers.ADAM
        options.probe_position_options.step_size = 1e-1
        
        options.opr_mode_weight_options.initial_weights = np.array([1, 1e-6, 1e-6, 1e-6])
        options.opr_mode_weight_options.optimize_intensity_variation = True
        options.opr_mode_weight_options.optimizable = True
        
        options.reconstructor_options.num_epochs = 4
        options.reconstructor_options.batch_size = 40
        options.reconstructor_options.default_device = api.Devices.GPU
        options.reconstructor_options.displayed_loss_function = api.LossFunctions.MSE_SQRT
        options.reconstructor_options.allow_nondeterministic_algorithms = False
        
        task = PtychographyTask(options)
        task.run()
            
        recon = task.get_data_to_cpu(name='object', as_numpy=True)[0]
        return recon
    
    
if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--generate-gold', action='store_true')
    args = parser.parse_args()

    tester = Test2dPtychoOptPlan()
    tester.setup_method(name="", generate_data=False, generate_gold=args.generate_gold, debug=True)
    tester.test_2d_ptycho_opt_plan()



# Copyright © 2025 UChicago Argonne, LLC All right reserved
# Full license accessible at https://github.com//AdvancedPhotonSource/pty-chi/blob/main/LICENSE

from typing import Optional, TYPE_CHECKING, Tuple, Union
import logging

import torch
from torch.utils.data import Dataset
from torch import Tensor
from torch.utils.data import DataLoader
import math

import ptychi.api as api
from ptychi.reconstructors.base import (
    AnalyticalIterativePtychographyReconstructor,
    LossTracker,
)
from ptychi.api.options.dm import DMReconstructorOptions
from ptychi.timing.timer_utils import timer
import ptychi.image_proc as ip

if TYPE_CHECKING:
    import ptychi.data_structures.parameter_group as pg

logger = logging.getLogger(__name__)


class DMReconstructor(AnalyticalIterativePtychographyReconstructor):
    """
    Difference map algorithm reconstructor class

    ### On memory usage
    The difference map algorithm takes up more memory than algorithms like LSQML
    and PIE because the entire exit wave `self.psi` must be stored in memory.

    For the least amount of memory usage, set the `chunk_length` option to 1.
    `chunk_length` has no effect on the convergence.

    ### On batching
    The `batch_size` option is not used by this reconstructor.
    """

    parameter_group: "pg.PlanarPtychographyParameterGroup"

    def __init__(
        self,
        parameter_group: "pg.PlanarPtychographyParameterGroup",
        dataset: Dataset,
        options: Optional["api.options.dm.DMReconstructorOptions"] = None,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(
            parameter_group=parameter_group,
            dataset=dataset,
            options=options,
            *args,
            **kwargs,
        )
        self.forward_model.retain_intermediates = False
        self.options: DMReconstructorOptions = self.options

    def check_inputs(self, *args, **kwargs):
        if self.parameter_group.object.is_multislice:
            raise NotImplementedError("DMReconstructor only supports 2D objects.")
        if self.parameter_group.probe.has_multiple_opr_modes:
            raise NotImplementedError("DMReconstructor does not support multiple OPR modes yet.")
        if (
            self.parameter_group.probe_positions.position_correction.options.correction_type
            is not api.enums.PositionCorrectionTypes.GRADIENT
        ):
            raise NotImplementedError("DMReconstructor only supports gradient position correction.")
        if self.options.batch_size != DMReconstructorOptions.batch_size:
            logger.warning("Difference map reconstruction does not support batching!")

    def build_loss_tracker(self):
        if self.options.displayed_loss_function is not None:
            logger.warning(
                "The loss tracker is hard-coded to record the DM error. "
                "The specified metric function will not be used!"
            )
        self.loss_tracker = LossTracker(metric_function=None)

    def build_dataloader(self):
        """
        Build the dataloader for the DM reconstructor. There is no
        batching in this reconstructor. Each batch is always a full,
        sequential sampling of the dataset.
        """
        data_loader_kwargs = {
            "dataset": self.dataset,
            "generator": torch.Generator(device=torch.get_default_device()),
            "batch_size": self.parameter_group.probe_positions.n_scan_points,
            "shuffle": False,
        }
        self.dataloader = DataLoader(**data_loader_kwargs)
        self.dataset.move_attributes_to_device(torch.get_default_device())

    @timer()
    def run_minibatch(self, input_data, y_true, *args, **kwargs):
        dm_error_squared = self.compute_updates(y_true, self.dataset.valid_pixel_mask)
        self.loss_tracker.update_batch_loss(loss=dm_error_squared.sqrt())

    @timer()
    def compute_updates(self, y_true: Tensor, valid_pixel_mask: Tensor) -> Tensor:
        """
        Compute the updates to the object, probe, and exit wave using the procedure
        described here: [Probe retrieval in ptychographic coherent diffractive imaging
        ](https://doi.org/10.1016/j.ultramic.2008.12.011).
        """

        object_ = self.parameter_group.object
        probe = self.parameter_group.probe
        probe_positions = self.parameter_group.probe_positions

        # Get indices used for dividing data into chunks
        start_pts = list(range(0, probe_positions.n_scan_points, self.options.chunk_length))
        end_pts = list(
            range(
                self.options.chunk_length,
                probe_positions.n_scan_points + self.options.chunk_length,
                self.options.chunk_length,
            )
        )
        end_pts[-1] = min(end_pts[-1], probe_positions.n_scan_points)
        n_chunks = math.ceil(probe_positions.n_scan_points / self.options.chunk_length)

        # Initialize the exit wave
        if self.current_epoch == 0:
            self.psi = torch.zeros(
                (
                    probe_positions.n_scan_points,
                    probe.n_modes,
                    *probe.get_spatial_shape(),
                ),
                dtype=object_.data.dtype,
            )
            for i in range(n_chunks):
                self.psi[start_pts[i] : end_pts[i]] = self.calculate_exit_wave_chunk(
                    start_pts[i], end_pts[i]
                )

        # Calculate the dm exit wave and probe update in chunks
        probe_numerator = torch.zeros_like(probe.get_opr_mode(0))
        probe_denominator = torch.zeros_like(probe.get_opr_mode(0).abs())
        delta_pos = torch.zeros_like(probe_positions.data)
        dm_error_squared = 0
        for i in range(n_chunks):
            obj_patches, dm_error_squared, new_psi = self.apply_dm_update_to_exit_wave_chunk(
                start_pts[i], end_pts[i], y_true, valid_pixel_mask, dm_error_squared
            )
            if probe.optimization_enabled(self.current_epoch):
                self.add_to_probe_update_terms(
                    probe_numerator,
                    probe_denominator,
                    obj_patches,
                    start_pts[i],
                    end_pts[i],
                )
            if probe_positions.optimization_enabled(self.current_epoch):
                delta_pos[start_pts[i] : end_pts[i]] = self.get_positions_update_chunk(
                    indices=torch.Tensor(list(range(start_pts[i], end_pts[i]))).to(int),
                    obj_patches=obj_patches,
                    chi=self.psi[start_pts[i] : end_pts[i]] - new_psi,
                )

        # Update the probe
        if probe.optimization_enabled(self.current_epoch):
            probe_update = self.calculate_probe_update(probe_numerator, probe_denominator)
            self.update_probe(probe_update)

        # Update the object
        if object_.optimization_enabled(self.current_epoch):
            updated_object = self.calculate_object_update(start_pts, end_pts)
            self.update_object(updated_object)

        if probe_positions.optimization_enabled(self.current_epoch):
            probe_positions.set_grad(-delta_pos)
            probe_positions.step_optimizer()

        return dm_error_squared

    @timer()
    def calculate_probe_update(self, probe_numerator: Tensor, probe_denominator: Tensor) -> Tensor:
        probe = self.parameter_group.probe
        probe_update = probe_numerator / torch.sqrt(
            probe_denominator**2 + (0.05 * probe_denominator.max()) ** 2
        )
        probe_update = (
            probe.options.inertia * probe.data + (1 - probe.options.inertia) * probe_update
        )
        return probe_update

    @timer()
    def update_probe(self, probe_update: Tensor):
        probe = self.parameter_group.probe
        probe.set_data(probe_update)

    @timer()
    def calculate_exit_wave_chunk(
        self, start_pt: int, end_pt: int, return_obj_patches: bool = False
    ) -> Union[Tensor, Tuple[Tensor, Tensor]]:
        object_ = self.parameter_group.object
        probe = self.parameter_group.probe
        positions = self.parameter_group.probe_positions.tensor

        obj_patches = object_.extract_patches(
            positions[start_pt:end_pt].round().int(), probe.get_spatial_shape(), integer_mode=True
        )
        psi = self.forward_model.forward_real_space(
            indices=torch.arange(start_pt, end_pt, device=obj_patches.device).long(),
            obj_patches=obj_patches,
        )

        if return_obj_patches:
            return psi, obj_patches
        else:
            return psi

    @timer()
    def apply_dm_update_to_exit_wave_chunk(
        self,
        start_pt: int,
        end_pt: int,
        y_true: Tensor,
        valid_pixel_mask: Tensor,
        dm_error_squared: Tensor,
    ) -> Tuple[Tensor, Tensor, Tensor]:
        """
        Calculate the updated exit wave according to equation (9) in
        this paper: https://doi.org/10.1016/j.ultramic.2008.12.011
        """
        # Mappings between this code and the paper:
        # - self.psi --> psi_n
        # - revised_psi --> 2 * Pi_o(psi_n)- psi_n
        # - new_psi --> Pi_o(psi_n)

        probe = self.parameter_group.probe

        # Get the update exit wave
        new_psi, obj_patches = self.calculate_exit_wave_chunk(
            start_pt, end_pt, return_obj_patches=True
        )
        # Propagate to detector plane
        revised_psi = self.forward_model.free_space_propagator.propagate_forward(
            2 * new_psi - self.psi[start_pt:end_pt]
        )
        # Replace intensities
        revised_psi = torch.where(
            valid_pixel_mask.repeat(revised_psi.shape[0], probe.n_modes, 1, 1),
            self.replace_propagated_exit_wave_magnitude(revised_psi, y_true[start_pt:end_pt]),
            revised_psi,
        )
        # Propagate back to sample plane
        revised_psi = self.forward_model.free_space_propagator.propagate_backward(revised_psi)
        # Update the exit wave
        psi_update = (revised_psi - new_psi) * self.options.exit_wave_update_relaxation
        self.psi[start_pt:end_pt] += psi_update

        dm_error_squared += (psi_update.abs() ** 2).sum()

        return obj_patches, dm_error_squared, new_psi

    @timer()
    def add_to_probe_update_terms(
        self,
        probe_numerator: Tensor,
        probe_denominator: Tensor,
        obj_patches: Tensor,
        start_pt: int,
        end_pt: int,
    ):
        "Add to the running totals for the probe update numerator and denominator"
        indices = torch.arange(start_pt, end_pt, device=probe_numerator.device).long()
        numerator_update = obj_patches.conj() * self.psi[start_pt:end_pt]
        denominator_update = obj_patches.abs() ** 2
        # Before summing in the batch dimension, the updates should be adjointly shifted to backpropagate
        # through the subpixel shifts of the probe.
        numerator_update = self.adjoint_shift_probe_update_direction(
            indices, numerator_update, first_mode_only=True
        )
        denominator_update = self.adjoint_shift_probe_update_direction(
            indices, denominator_update, first_mode_only=True
        )
        numerator_update = numerator_update.sum(0)
        denominator_update = denominator_update.sum(0)
        probe_numerator += numerator_update
        probe_denominator += denominator_update

    @timer()
    def calculate_object_update(self, start_pts: list[int], end_pts: list[int]) -> Tensor:
        "Calculate and apply the object update."

        object_ = self.parameter_group.object
        positions = self.parameter_group.probe_positions.tensor

        # Calculate object update
        # Iterating over the object numerator calculation is more memory efficient
        object_numerator = torch.zeros_like(object_.get_slice(0))
        object_denominator = torch.zeros_like(object_.get_slice(0), dtype=positions.dtype)
        for i in range(len(start_pts)):
            indices = torch.arange(start_pts[i], end_pts[i], device=object_numerator.device).long()
            p = self.forward_model.get_unique_probes(indices, always_return_probe_batch=True)
            p = self.forward_model.shift_unique_probes(indices, p, first_mode_only=True)

            object_numerator = ip.place_patches_integer(
                object_numerator,
                positions[start_pts[i] : end_pts[i]].round().int() + object_.pos_origin_coords,
                patches=(p.conj() * self.psi[start_pts[i] : end_pts[i]]).sum(1),
                op="add",
            )

            object_denominator = ip.place_patches_integer(
                object_denominator,
                positions[start_pts[i] : end_pts[i]].round().int() + object_.pos_origin_coords,
                patches=(p.abs() ** 2).sum(1),
                op="add",
            )

        # Calculate DM object update
        object_update = object_numerator / torch.sqrt(
            object_denominator**2 + (0.05 * object_denominator.max()) ** 2
        )
        # Apply inertia
        object_update = object_.get_slice(0) * object_.options.inertia + object_update * (
            1 - object_.options.inertia
        )
        # Clamp the object amplitude -- this is rarely needed
        idx = object_update.abs() > object_.options.amplitude_clamp_limit
        object_update[idx] = (
            object_update[idx] / (object_update[idx].abs())
        ) * object_.options.amplitude_clamp_limit

        return object_update

    @timer()
    def update_object(self, object_update: Tensor):
        # Apply object update
        self.parameter_group.object.set_data(object_update)

    @timer()
    def get_positions_update_chunk(
        self, indices: Tensor, obj_patches: Tensor, chi: Tensor
    ) -> Tensor:
        "Calculate the probe position update. Only gradient based updates are allowed for now."

        probe_positions = self.parameter_group.probe_positions
        probe = self.forward_model.get_unique_probes(indices, always_return_probe_batch=True)
        probe = self.forward_model.shift_unique_probes(indices, probe, first_mode_only=True)

        delta_pos = probe_positions.position_correction.get_update(
            chi,
            obj_patches,
            None,
            probe,
            None,
        )

        return delta_pos



# Copyright © 2025 UChicago Argonne, LLC All right reserved
# Full license accessible at https://github.com//AdvancedPhotonSource/pty-chi/blob/main/LICENSE

import dataclasses
from enum import StrEnum, auto
from typing import Sequence, Optional
import cv2


class MovieFileTypes(StrEnum):
    GIF = auto()
    MP4 = auto()


class MovieSubjectTypes(StrEnum):
    OBJECT = auto()
    PROBE = auto()


class ProcessFunctionType(StrEnum):
    PHASE = auto()
    MAGNITUDE = auto()


class ProbePlotTypes(StrEnum):
    INCOHERENT_SUM = auto()
    SEPERATE_MODES = auto()


@dataclasses.dataclass
class SnapshotSettings:
    stride: int = 1
    "The number of epochs between each frame acquisition"

    scale: int = 4
    "How much each acquired frame is downsampled by"


@dataclasses.dataclass
class MovieFileSettings:
    "Settings related to look and playback of the movie file."

    fps: int = 5
    "Frames per second in the saved movie file"

    high_contrast: bool = False

    colormap: int = cv2.COLORMAP_BONE

    compress: bool = True

    upper_bound: Optional[float] = None

    lower_bound: Optional[float] = None


@dataclasses.dataclass
class ObjectMovieSettings:
    process_function: ProcessFunctionType = ProcessFunctionType.PHASE

    slice_index: int = 0
    "Index of the object to select"

    snapshot: SnapshotSettings = dataclasses.field(default_factory=SnapshotSettings)

    movie_file: MovieFileSettings = dataclasses.field(default_factory=MovieFileSettings)

    save_intermediate_data_to_hdf5: bool = False


@dataclasses.dataclass
class ProbeMovieSettings:
    process_function: ProcessFunctionType = ProcessFunctionType.MAGNITUDE

    plot_type: ProbePlotTypes = ProbePlotTypes.INCOHERENT_SUM
    """
    Determines what type of plot to make for the probe. You can choose 
    between plotting the total intensity (INCOHERENT_SUM) and plotting
    each mode seperately (SEPERATE_MODES).
    """

    mode_indices: Optional[Sequence[int]] = None
    """
    Indices of modes to be selected for plotting. This is only used 
    when `plot_type` is `SEPERATE_MODES`.
    """

    snapshot: SnapshotSettings = dataclasses.field(default_factory=SnapshotSettings)

    movie_file: MovieFileSettings = dataclasses.field(default_factory=MovieFileSettings)

    save_intermediate_data_to_hdf5: bool = False



# Copyright © 2025 UChicago Argonne, LLC All right reserved
# Full license accessible at https://github.com//AdvancedPhotonSource/pty-chi/blob/main/LICENSE

from typing import Optional, Tuple, Sequence, TYPE_CHECKING
import logging

import pandas as pd
import torch
from torch import Tensor
from torch.utils.data import DataLoader, Dataset
import tqdm
from ptychi.timing.timer_utils import timer
import ptychi.movies as movies

from ptychi.utils import to_numpy
import ptychi.maps as maps
import ptychi.forward_models as fm
import ptychi.api.enums as enums
import ptychi.io_handles as io
import ptychi.image_proc as ip
if TYPE_CHECKING:
    import ptychi.data_structures.parameter_group as pg
    import ptychi.api as api

logger = logging.getLogger(__name__)


class LossTracker:
    def __init__(
        self,
        metric_function: Optional[torch.nn.Module] = None,
        always_compute_loss: bool = False,
        *args,
        **kwargs,
    ) -> None:
        """
        Parameters
        ----------
        metric_function : callable, optional
            A function that takes y_pred and y_true and returns a loss.
        always_compute_loss : bool, optional
            Determines the behavior of update_batch_loss. If True,
            the loss is computed using the metric function as long as y_pred and y_true
            are given. Otherwise, the tracker logs the provided loss value if it is given,
            only computing the loss when it is not.
        """
        super().__init__(*args, **kwargs)
        self.table = pd.DataFrame(columns=["epoch", "loss", "reg_loss"])
        self.table["epoch"] = self.table["epoch"].astype(int)
        self.metric_function = metric_function
        self.epoch_loss = 0.0
        self.accumulated_num_batches = 0
        self.epoch_reg_loss = 0.0
        self.accumulated_num_reg_batches = 0
        self.epoch = 0
        self.always_compute_loss = always_compute_loss

    def conclude_epoch(self, epoch: Optional[int] = None) -> None:
        self.epoch_loss = self.epoch_loss / self.accumulated_num_batches
        self.epoch_reg_loss = self.epoch_reg_loss / max(self.accumulated_num_reg_batches, 1)
        if epoch is None:
            epoch = self.epoch
            self.epoch += 1
        else:
            self.epoch = epoch + 1
        self.table.loc[len(self.table)] = [epoch, self.epoch_loss, self.epoch_reg_loss]
        self.epoch_loss = 0.0
        self.epoch_reg_loss = 0.0
        self.accumulated_num_batches = 0
        self.accumulated_num_reg_batches = 0

    def update_batch_loss(
        self,
        y_pred: Optional[torch.Tensor] = None,
        y_true: Optional[torch.Tensor] = None,
        loss: Optional[float] = None,
    ) -> None:
        """
        Update the loss after processing a minibatch. This routine decides whether
        to compute the loss using the metric function or a provided loss value:

        - If always_compute_loss is True, the loss is always computed using the metric
          function. An exception is thrown if data (y_pred and y_true) are not provided,
          or if the metric function is not provided.
        - If always_compute_loss is False, the loss is updated using the provided loss
          as long as it is provided.
        - Otherwise, the loss is computed using the metric function and provided data.

        At least one of (y_pred, y_true) and loss must be provided. Also, in the former
        case, metric_function must be provided.

        Parameters
        ----------
        y_pred : Optional[Tensor]
            The predicted or simulated data.
        y_true : Optional[Tensor]
            The measured or labeled data.
        loss : Optional[float]
            The precalculated loss value.
        """
        data_provided = (
            y_pred is not None and y_true is not None and self.metric_function is not None
        )
        loss_provided = loss is not None
        if self.always_compute_loss:
            if not data_provided:
                raise ValueError(
                    "Always_compute_loss requires (y_pred, y_true) and metric_function to be provided."
                )
        if not (data_provided or loss_provided):
            raise ValueError(
                "One of (y_pred, y_true) and (loss,) must be provided. Also, in the former case, metric_function must be provided."
            )

        if loss_provided and not self.always_compute_loss:
            self.update_batch_loss_with_value(loss)
        else:
            self.update_batch_loss_with_metric_function(y_pred, y_true)

    def update_batch_loss_with_metric_function(
        self, y_pred: torch.Tensor, y_true: torch.Tensor
    ) -> None:
        if self.metric_function is None:
            raise ValueError("update_batch_loss_with_metric_function requires a metric function.")
        batch_loss = self.metric_function(y_pred, y_true)
        batch_loss = to_numpy(batch_loss)
        self.epoch_loss = self.epoch_loss + batch_loss
        self.accumulated_num_batches = self.accumulated_num_batches + 1

    def update_batch_loss_with_value(self, loss: float) -> None:
        loss = to_numpy(loss)
        self.epoch_loss = self.epoch_loss + loss
        self.accumulated_num_batches = self.accumulated_num_batches + 1
        
    def update_batch_regularization_loss(self, reg_loss: float) -> None:
        reg_loss = to_numpy(reg_loss)
        self.epoch_reg_loss = self.epoch_reg_loss + reg_loss
        self.accumulated_num_reg_batches = self.accumulated_num_reg_batches + 1

    def print(self) -> None:
        print(self.table)

    def print_latest(self) -> None:
        logger.info(
            "Epoch: {}, Loss: {}".format(int(self.table.iloc[-1].epoch), self.table.iloc[-1].loss)
        )

    def to_csv(self, path: str) -> None:
        self.table.to_csv(path, index=False)


class Reconstructor:
    def __init__(
        self,
        parameter_group: "pg.ParameterGroup",
        options: Optional["api.base.ReconstructorOptions"] = None,
    ) -> None:
        self.loss_tracker = LossTracker()
        self.parameter_group = parameter_group

        if options is None:
            options = self.get_option_class()()
        self.options = options

    def check_inputs(self, *args, **kwargs):
        pass

    def build(self) -> None:
        self.check_inputs()

    def get_option_class(self):
        try:
            return self.__class__.__init__.__annotations__["options"]
        except KeyError:
            return api.options.base.ReconstructorOptions

    def get_config_dict(self) -> dict:
        d = self.parameter_group.get_config_dict()
        reconstructor_options = {"name": self.__class__.__name__}
        reconstructor_options.update(self.options.__dict__)
        d["reconstructor_options"] = reconstructor_options
        return d


class PtychographyReconstructor(Reconstructor):
    parameter_group: "pg.PtychographyParameterGroup"

    def __init__(
        self,
        parameter_group: "pg.PtychographyParameterGroup",
        options: Optional["api.base.ReconstructorOptions"] = None,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(parameter_group, options=options, *args, **kwargs)
        
        self.parameter_group.object.update_pos_origin_coordinates(
            self.parameter_group.probe_positions.data
        )
        self.parameter_group.object.build_roi_bounding_box(
            self.parameter_group.probe_positions
        )


class IterativeReconstructor(Reconstructor):
    def __init__(
        self,
        parameter_group: "pg.ParameterGroup",
        dataset: Dataset,
        options: Optional["api.base.ReconstructorOptions"] = None,
        *args,
        **kwargs,
    ) -> None:
        """
        Parameters
        ----------
        parameter_group : ds.ParameterGroup
            The ParameterGroup containing optimizable and non-optimizable parameters.
        dataset : Dataset
            The dataset containing diffraction patterns.
        options : ReconstructorOptions
            The options for the reconstruction.
        """
        super().__init__(parameter_group, options=options, *args, **kwargs)
        self.batch_size = options.batch_size
        self.dataset = dataset
        self.n_epochs = options.num_epochs
        self.dataloader = None
        self.displayed_loss_function = options.displayed_loss_function
        if self.displayed_loss_function is not None:
            self.displayed_loss_function = maps.get_loss_function_by_enum(
                options.displayed_loss_function
            )()
        self.current_epoch = 0
        self.current_minibatch = 0

    def build(self) -> None:
        super().build()
        self.build_dataloader()
        self.build_loss_tracker()
        self.build_counter()

    def build_dataloader(self, batch_sampler=None):
        data_loader_kwargs = {
            "dataset": self.dataset,
            "generator": torch.Generator(device=torch.get_default_device()),
        }
        if batch_sampler is not None:
            data_loader_kwargs["batch_sampler"] = batch_sampler
        else:
            data_loader_kwargs["batch_size"] = self.batch_size
            data_loader_kwargs["shuffle"] = True

        self.dataloader = DataLoader(**data_loader_kwargs)
        self.dataset.move_attributes_to_device(torch.get_default_device())

    def build_loss_tracker(self):
        self.loss_tracker = LossTracker(metric_function=self.displayed_loss_function)

    def build_counter(self):
        self.pbar = tqdm.tqdm(
            total=self.options.num_epochs, 
            disable=logger.level > logging.INFO, 
            leave=False
        )
        self.current_epoch = 0

    def get_config_dict(self) -> dict:
        d = super().get_config_dict()
        d.update({"batch_size": self.batch_size, "n_epochs": self.n_epochs})
        return d
    
    def prepare_batch_data(self, batch_data: Sequence[Tensor]) -> Tuple[Sequence[Tensor], Tensor]:
        input_data = batch_data[:-1]
        
        if input_data[0].ndim != 1:
            raise ValueError(
                "The first returned value of the dataset should be a 1D tensor "
                "giving the indices, but got a tensor of shape {}.".format(input_data[0].shape)
            )
            
        # Put input data on CPU so that DataParallel can distribute them later.
        input_data = [x.cpu() for x in input_data]
        
        # Put measured data on device.
        y_true = batch_data[-1]
        if y_true.device.type != torch.get_default_device().type:
            y_true = y_true.to(torch.get_default_device())
        return input_data, y_true

    def run_minibatch(self, input_data: Sequence[Tensor], y_true: Tensor, *args, **kwargs) -> None:
        """
        Process batch, update parameters, calculate loss, and update loss tracker.

        Parameters
        ----------
        input_data : list of Tensor
            A list of input data. In many cases it is [indices].
        y_true : Tensor
            The measured data.
        """
        raise NotImplementedError

    def run_pre_run_hooks(self) -> None:
        pass

    def run_pre_update_hooks(self) -> None:
        pass

    def run_pre_epoch_hooks(self) -> None:
        pass

    def run_post_update_hooks(self) -> None:
        pass

    def run_post_epoch_hooks(self) -> None:
        pass

    @timer()
    def run(self, n_epochs: Optional[int] = None, *args, **kwargs):
        if self.current_epoch == 0:
            self.run_pre_run_hooks()

        n_epochs = n_epochs if n_epochs is not None else self.n_epochs
        for _ in range(n_epochs):
            self.run_pre_epoch_hooks()
            self.current_minibatch = 0
            for batch_data in self.dataloader:
                input_data, y_true = self.prepare_batch_data(batch_data)
                self.run_pre_update_hooks()
                self.run_minibatch(input_data, y_true)
                self.run_post_update_hooks()
                self.current_minibatch += 1
            self.run_post_epoch_hooks()
            self.loss_tracker.conclude_epoch(epoch=self.current_epoch)
            self.loss_tracker.print_latest()

            if movies.MOVIES_INSTALLED:
                movies.api.update_movie_builders(self)

            self.current_epoch += 1
            self.pbar.update(1)


class IterativePtychographyReconstructor(IterativeReconstructor, PtychographyReconstructor):
    parameter_group: "pg.PtychographyParameterGroup"

    def __init__(
        self,
        parameter_group: "pg.PtychographyParameterGroup",
        options: Optional["api.base.ReconstructorOptions"] = None,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(parameter_group, options=options, *args, **kwargs)

    def build_dataloader(self):
        batch_sampler = None
        if self.options.batching_mode == enums.BatchingModes.COMPACT:
            batch_sampler = io.PtychographyCompactBatchSampler(
                self.parameter_group.probe_positions.data.cpu(), self.batch_size
            )
        elif self.options.batching_mode == enums.BatchingModes.UNIFORM:
            batch_sampler = io.PtychographyUniformBatchSampler(
                self.parameter_group.probe_positions.data.cpu(), self.batch_size
            )
        return super().build_dataloader(batch_sampler=batch_sampler)

    def update_preconditioners(self):
        # Update preconditioner of the object only if:
        # - the preconditioner does not exist, or
        # - it is within the 10 epochs after the probe starts being optimized, or
        # - it has been 10 epochs since the preconditioner was last updated.
        if self.parameter_group.object.preconditioner is None or (
            (self.current_epoch > self.parameter_group.probe.optimization_plan.start)
            and (
                (self.current_epoch - self.parameter_group.probe.optimization_plan.start < 10)
                or (
                    (self.current_epoch - self.parameter_group.probe.optimization_plan.start) % 10
                    == 0
                )
            )
        ):
            self.parameter_group.object.update_preconditioner(
                probe=self.parameter_group.probe,
                probe_positions=self.parameter_group.probe_positions,
                patterns=self.dataset.patterns,
                use_all_modes=self.parameter_group.object.options.build_preconditioner_with_all_modes
            )
            
    def run_pre_epoch_hooks(self) -> None:
        with torch.no_grad():
            self.update_preconditioners()
            

    def run_post_epoch_hooks(self) -> None:
        with torch.no_grad():
            probe = self.parameter_group.probe
            object_ = self.parameter_group.object
            positions = self.parameter_group.probe_positions

            # Apply probe power constraint.
            if probe.options.power_constraint.is_enabled_on_this_epoch(self.current_epoch):
                probe.constrain_probe_power(
                    self.parameter_group.object, self.parameter_group.opr_mode_weights
                )

            # Apply incoherent mode orthogonality constraint.
            if probe.options.orthogonalize_incoherent_modes.is_enabled_on_this_epoch(
                self.current_epoch
            ):
                probe.constrain_incoherent_modes_orthogonality()

            # Apply OPR orthogonality constraint.
            if probe.options.orthogonalize_opr_modes.is_enabled_on_this_epoch(self.current_epoch):
                probe.constrain_opr_mode_orthogonality(
                    self.parameter_group.opr_mode_weights, update_weights_in_place=True
                )

            # Regularize multislice reconstruction.
            if object_.options.multislice_regularization.is_enabled_on_this_epoch(
                self.current_epoch
            ):
                object_.regularize_multislice()

            # Apply smoothness constraint.
            if object_.options.smoothness_constraint.is_enabled_on_this_epoch(self.current_epoch):
                object_.constrain_smoothness()

            # Remove grid artifacts.
            if object_.options.remove_grid_artifacts.is_enabled_on_this_epoch(self.current_epoch):
                object_.remove_grid_artifacts()

            # Apply position constraint.
            if positions.position_mean_constraint_enabled(self.current_epoch):
                positions.constrain_position_mean()

            # Update compact mode clustering.
            if (
                self.options.batching_mode == enums.BatchingModes.COMPACT
                and self.options.compact_mode_update_clustering
                and positions.optimization_enabled(self.current_epoch)
                and (self.current_epoch - positions.optimization_plan.start)
                % self.options.compact_mode_update_clustering_stride
                == 0
            ):
                self.dataloader.batch_sampler.update_clusters(positions.data.detach().cpu())
                
            # Apply probe support constraint.
            if probe.options.support_constraint.is_enabled_on_this_epoch(self.current_epoch):
                probe.constrain_support()
                
            # Apply probe center constraint.
            if probe.options.center_constraint.is_enabled_on_this_epoch(self.current_epoch):
                probe.center_probe()

            # Remove object-probe ambiguity.
            if object_.options.remove_object_probe_ambiguity.is_enabled_on_this_epoch(self.current_epoch):
                object_.remove_object_probe_ambiguity(probe, update_probe_in_place=True)
            
            # Smooth OPR weights.
            opr_mode_weights = self.parameter_group.opr_mode_weights
            if opr_mode_weights.options.smoothing.is_enabled_on_this_epoch(self.current_epoch):
                opr_mode_weights.smooth_weights()
            opr_mode_weights.remove_outliers()
            
            # Position affine transformation constraint.
            if (
                positions.options.affine_transform_constraint.is_position_weight_update_enabled_on_this_epoch(
                    self.current_epoch
                )
            ):
                positions.update_position_weights(probe, object_, self.options.batch_size)
            if positions.options.affine_transform_constraint.is_enabled_on_this_epoch(self.current_epoch):
                positions.update_affine_transform_matrix()
                if positions.options.affine_transform_constraint.apply_constraint:
                    positions.apply_affine_transform_constraint()


class AnalyticalIterativeReconstructor(IterativeReconstructor):
    def __init__(
        self,
        parameter_group: "pg.ParameterGroup",
        dataset: Dataset,
        options: Optional["api.base.ReconstructorOptions"] = None,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(
            parameter_group=parameter_group,
            dataset=dataset,
            options=options,
            *args,
            **kwargs,
        )
        self.update_step_module: torch.nn.Module = None

    def build(self) -> None:
        super().build()
        self.build_update_step_module()

    def build_update_step_module(self, *args, **kwargs):
        update_step_func = self.compute_updates
        par_group = self.parameter_group

        class EncapsulatedUpdateStep(torch.nn.Module):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)
                self.parameter_module_dict = torch.nn.ModuleDict(par_group.__dict__)

            def forward(self, *args, **kwargs):
                return update_step_func(self, *args, **kwargs)

            def process_updates(self, *args):
                ret = []
                for v in args:
                    if v is not None:
                        v = v.unsqueeze(0)
                        if v.is_complex():
                            v = torch.stack([v.real, v.imag], dim=-1)
                    ret.append(v)
                return tuple(ret)

        self.update_step_module = EncapsulatedUpdateStep()
        if not torch.get_default_device().type == "cpu":
            # TODO: use CUDA stream instead of DataParallel for non-AD reconstructor.
            # https://poe.com/s/NZUVScEEGLxBE5ZDmKc0
            self.update_step_module = torch.nn.DataParallel(self.update_step_module)
            self.update_step_module.to(torch.get_default_device())

    @staticmethod
    def compute_updates(
        update_step_module: torch.nn.Module, *args, **kwargs
    ) -> Tuple[Tuple[Tensor, ...], float]:
        """
        Calculate the update vectors of optimizable parameters that should be
        applied later.

        Parameters
        ----------
        update_step_module : torch.nn.Module
            The module that contains the update step function.
        *args
            Passed to the update step function.
        **kwargs
            Passed to the update step function.

        Returns
        -------
        updates : tuple of torch.Tensor
            The update vectors of optimizable parameters. The shape of each update
            vector is [n_replica, ..., (2 if complex else none)]. `n_replica` is 1 if
            only CPU is used, otherwise it is the number of GPUs. If the update vector
            is complex, it should be returned as real tensor with its real and
            imaginary parts concatenated at the last dimension.
        loss : float
            The batch loss.
        """
        raise NotImplementedError

    def apply_updates(self, *args, **kwargs):
        raise NotImplementedError

    @timer()
    def run(self, n_epochs: Optional[int] = None, *args, **kwargs):
        with torch.no_grad():
            return super().run(n_epochs=n_epochs, *args, **kwargs)


class AnalyticalIterativePtychographyReconstructor(
    AnalyticalIterativeReconstructor, IterativePtychographyReconstructor
):
    parameter_group: "pg.PtychographyParameterGroup"

    def __init__(
        self,
        parameter_group: "pg.PtychographyParameterGroup",
        dataset: Dataset,
        options: Optional["api.base.ReconstructorOptions"] = None,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(
            parameter_group=parameter_group,
            dataset=dataset,
            options=options,
            *args,
            **kwargs,
        )
        self.forward_model = None
        self.build_forward_model()

    def build_forward_model(self):
        self.forward_model = fm.PlanarPtychographyForwardModel(
            parameter_group=self.parameter_group, 
            retain_intermediates=True,
            detector_size=tuple(self.dataset.patterns.shape[-2:]),
            wavelength_m=self.dataset.wavelength_m,
            free_space_propagation_distance_m=self.dataset.free_space_propagation_distance_m,
            pad_for_shift=self.options.forward_model_options.pad_for_shift,
            low_memory_mode=self.options.forward_model_options.low_memory_mode,
        )

    def run_post_epoch_hooks(self) -> None:
        with torch.no_grad():
            super().run_post_epoch_hooks()

            object_ = self.parameter_group.object

            # Apply object L1-norm constraint.
            if object_.options.l1_norm_constraint.is_enabled_on_this_epoch(self.current_epoch):
                object_.constrain_l1_norm()
                
            # Apply object L2-norm constraint.
            if object_.options.l2_norm_constraint.is_enabled_on_this_epoch(self.current_epoch):
                object_.constrain_l2_norm()
                
            # Apply total variation constraint.
            if object_.options.total_variation.is_enabled_on_this_epoch(self.current_epoch):
                object_.constrain_total_variation()

    def run_pre_run_hooks(self) -> None:
        self.prepare_data()

    def prepare_data(self, *args, **kwargs):
        self.parameter_group.probe.normalize_eigenmodes()
        logger.info("Probe eigenmodes normalized.")

    @staticmethod
    def replace_propagated_exit_wave_magnitude(
        psi: Tensor, actual_pattern_intensity: Tensor
    ) -> Tensor:
        """
        Replace the propogated exit wave amplitude.

        Parameters
        ----------
        psi : Tensor
            Predicted exit wave propagated to the detector plane.
        actual_pattern_intensity : Tensor
            The measured diffraction pattern at the detector.

        Returns
        -------
        Tensor
            Predicted exit wave with the phase from `psi` and magnitude equal to the square root
            of `actual_pattern_intensity`.
        """

        return (
            psi
            / ((psi.abs() ** 2).sum(1, keepdims=True).sqrt() + 1e-7)
            * torch.sqrt(actual_pattern_intensity + 1e-7)[:, None]
        )
        
    @timer()
    def adjoint_shift_probe_update_direction(self, indices, delta_p, first_mode_only=False):
        """
        Apply the adjoint operator of the subpixel shift to the probe done in forward model.
        
        Parameters
        ----------
        indices : Tensor
            Indices of diffraction patterns in the current batch.
        delta_p : Tensor
            A (batch_size, n_probe_modes, h, w) tensor giving the probe update direction.
        first_mode_only : bool
            If True, only the first mode is shifted.

        Returns
        -------
        Tensor
            A (batch_size, n_probe_modes, h, w) tensor giving the shifted probe update direction.
        """
        pos = self.parameter_group.probe_positions.data[indices]
        orig_shape = delta_p.shape
        fractional_shifts = pos - pos.round()
        
        if first_mode_only:
            delta_p_to_shift = delta_p[..., 0, :, :]
        else:
            delta_p_to_shift = delta_p.reshape(delta_p.shape[0] * delta_p.shape[1], *delta_p.shape[2:])
        
        delta_p_shifted = ip.shift_images(
            delta_p_to_shift, 
            fractional_shifts, 
            method=self.parameter_group.object.options.patch_interpolation_method,
            adjoint=True
        )
        
        if first_mode_only:
            delta_p = torch.cat([delta_p_shifted[..., None, :, :], delta_p[..., 1:, :, :]], dim=-3)
        else:
            delta_p = delta_p_shifted.reshape(orig_shape)
        return delta_p



# Copyright © 2025 UChicago Argonne, LLC All right reserved
# Full license accessible at https://github.com//AdvancedPhotonSource/pty-chi/blob/main/LICENSE

import dataclasses
from dataclasses import field

import ptychi.api.options.base as base
import ptychi.api.options.task as task_options
import ptychi.api.enums as enums


@dataclasses.dataclass
class BHReconstructorOptions(base.ReconstructorOptions):
    method: str = "GD"
    """
    Reconstruction algorithm: 'GD' (Gradient Descent) or 'CG' (Conjugate Gradients). 
    Batch processing should be performed using the GD method, while CG is significantly 
    faster when using a single batch, i.e., when all positions are included in the update.
    """

    def get_reconstructor_type(self) -> enums.Reconstructors:
        return enums.Reconstructors.BH


@dataclasses.dataclass
class BHObjectOptions(base.ObjectOptions):
    pass


@dataclasses.dataclass
class BHProbeOptions(base.ProbeOptions):
    rho: float = 1
    """   
    Scaling factor for the probe variable relative to the object variable.
    It addresses numerical instabilities caused by disparities in gradient magnitudes between variables, 
    ensuring balanced convergence. The optimal value of rho is currently determined experimentally, 
    based on how closely the initial guess aligns with the final solution.
    For near-field ptychography, a typical value is 1, while for far-field ptychography, it is 0.1.
    """


@dataclasses.dataclass
class BHProbePositionOptions(base.ProbePositionOptions):
    rho: float = 0.1
    """   
    Scaling factor for the probe positions relative to the object variable.
    It addresses numerical instabilities caused by disparities in gradient magnitudes between variables, 
    ensuring balanced convergence. The optimal value of rho is currently determined experimentally, 
    based on how closely the initial guess aligns with the final solution.
    For near-field ptychography, a typical value is 0.1, while for far-field ptychography, it is 1.
    """


@dataclasses.dataclass
class BHOPRModeWeightsOptions(base.OPRModeWeightsOptions):
    pass


@dataclasses.dataclass
class BHOptions(task_options.PtychographyTaskOptions):
    reconstructor_options: BHReconstructorOptions = field(default_factory=BHReconstructorOptions)

    object_options: BHObjectOptions = field(default_factory=BHObjectOptions)

    probe_options: BHProbeOptions = field(default_factory=BHProbeOptions)

    probe_position_options: BHProbePositionOptions = field(default_factory=BHProbePositionOptions)

    opr_mode_weight_options: BHOPRModeWeightsOptions = field(
        default_factory=BHOPRModeWeightsOptions
    )



# Copyright © 2025 UChicago Argonne, LLC All right reserved
# Full license accessible at https://github.com//AdvancedPhotonSource/pty-chi/blob/main/LICENSE

from typing import Optional, TYPE_CHECKING
import logging
import math

import torch
from torch.utils.data import Dataset

from ptychi.reconstructors.base import (
    AnalyticalIterativePtychographyReconstructor,
    LossTracker,
)
import ptychi.forward_models as fm
import ptychi.maths as pmath
import ptychi.api.enums as enums
from ptychi.timing.timer_utils import timer
import ptychi.image_proc as ip

if TYPE_CHECKING:
    import ptychi.data_structures.parameter_group as pg
    import ptychi.api as api

logger = logging.getLogger(__name__)


class LSQMLReconstructor(AnalyticalIterativePtychographyReconstructor):
    """
    The least square maximum likelihood (LSQ-ML) algorithm described in

    Odstrčil, M., Menzel, A., & Guizar-Sicairos, M. (2018). Iterative
    least-squares solver for generalized maximum-likelihood ptychography.
    Optics Express, 26(3), 3108–3123. doi:10.1364/oe.26.003108

    This implementation uses automatic differentiation to get necessary gradients,
    but other steps, including the solving of the step size, are done analytically.
    """

    parameter_group: "pg.PlanarPtychographyParameterGroup"
    options: "api.LSQMLReconstructorOptions"

    def __init__(
        self,
        parameter_group: "pg.PlanarPtychographyParameterGroup",
        dataset: Dataset,
        options: Optional["api.LSQMLReconstructorOptions"] = None,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(
            parameter_group=parameter_group,
            dataset=dataset,
            options=options,
            *args,
            **kwargs,
        )

        noise_model_params = (
            {} if options.noise_model == "poisson" else {"sigma": options.gaussian_noise_std}
        )
        self.noise_model = {
            "gaussian": fm.PtychographyGaussianNoiseModel,
            "poisson": fm.PtychographyPoissonNoiseModel,
        }[options.noise_model](
            **noise_model_params, valid_pixel_mask=self.dataset.valid_pixel_mask.clone()
        )

        self.alpha_psi_far = 0.5
        self.alpha_psi_far_all_pos = None
        self.alpha_object_all_pos_all_slices = torch.ones(
            [self.parameter_group.probe_positions.shape[0], self.parameter_group.object.n_slices],
            device=torch.get_default_device(),
        )
        self.alpha_probe_all_pos = torch.ones(
            self.parameter_group.probe_positions.shape[0], device=torch.get_default_device()
        )

        self.indices = []

        self.object_momentum_params = {}
        self.probe_momentum_params = {}

        self.accumulated_true_intensity = 0
        self.accumulated_pred_intensity = 0

        # Fourier error for momentum acceleration.
        self.accumulated_fourier_error = 0.0
        self.fourier_errors = []

    def check_inputs(self, *args, **kwargs):
        if self.parameter_group.opr_mode_weights.optimizer is not None:
            logger.warning(
                "Selecting optimizer for OPRModeWeights is not supported for "
                "LSQMLReconstructor and will be disregarded."
            )
        if (
            self.parameter_group.opr_mode_weights.n_opr_modes > 1
            and self.parameter_group.opr_mode_weights.data[:, 1:].abs().max() < 1e-9
        ):
            raise ValueError(
                "Weights of eigenmodes (the second and following OPR modes) in LSQMLReconstructor "
                "should not be all zero, which can cause numerical instability!"
            )

    def build(self) -> None:
        super().build()
        self.build_cached_variables()
        self.build_noise_model()

    def build_loss_tracker(self):
        f = (
            self.noise_model.nll
            if self.displayed_loss_function is None
            else self.displayed_loss_function
        )
        self.loss_tracker = LossTracker(metric_function=f)

    def build_noise_model(self):
        self.noise_model = self.noise_model.to(torch.get_default_device())

    def build_cached_variables(self):
        self.alpha_psi_far_all_pos = torch.full(
            size=(self.parameter_group.probe_positions.shape[0],), fill_value=0.5
        )

    @timer()
    def get_psi_far_step_size(self, y_pred, y_true, indices, eps=1e-5):
        if isinstance(self.noise_model, fm.PtychographyGaussianNoiseModel):
            alpha = torch.tensor(0.5, device=y_pred.device)  # Eq. 16
        elif isinstance(self.noise_model, fm.PtychographyPoissonNoiseModel):
            # This implementation reproduces PtychoShelves (gradient_descent_xi_solver)
            # and is different from Eq. 17 of Odstrcil (2018).
            xi = 1 - y_true / (y_pred + eps)
            for _ in range(2):
                alpha_prev = self.alpha_psi_far_all_pos[indices].mean()
                alpha = (xi * (y_pred - y_true / (1 - alpha_prev * xi))).sum(-1).sum(-1)
                alpha = alpha / (xi**2 * y_pred).sum(-1).sum(-1)
                # Use previous step size as momentum.
                alpha = 0.5 * alpha_prev + 0.5 * alpha
                alpha = alpha.clamp(0, 1)
                self.alpha_psi_far_all_pos[indices] = alpha
            # Add perturbation.
            alpha = alpha + torch.randn(alpha.shape, device=alpha.device) * 1e-2
            self.alpha_psi_far_all_pos[indices] = alpha
            logger.debug("poisson alpha_psi_far: mean = {}".format(torch.mean(alpha)))
        return alpha

    @timer()
    def run_reciprocal_space_step(self, y_pred, y_true, indices):
        """
        Run step 1 of LSQ-ML, which updates `psi`.

        Returns
        -------
        psi_opt : Tensor
            A (batch_size, n_probe_modes, h, w) complex tensor.
        """
        # gradient as in Eq. 12a/b
        psi_far_0 = self.forward_model.intermediate_variables["psi_far"]
        dl_dpsi_far = self.noise_model.backward_to_psi_far(y_pred, y_true, psi_far_0)
        self.alpha_psi_far = self.get_psi_far_step_size(y_pred, y_true, indices)
        psi_far = psi_far_0 - self.alpha_psi_far.view(-1, 1, 1, 1) * dl_dpsi_far  # Eq. 14

        psi_opt = self.forward_model.free_space_propagator.propagate_backward(psi_far)
        return psi_opt

    @timer()
    def run_real_space_step(self, psi_opt, indices):
        """
        Run real space step of LSQ-ML, which updates the object, probe, and other variables
        using psi updated in the reciprocal space step and backpropagated to real space.

        Parameters
        ----------
        psi_opt : Tensor
            A (batch_size, n_probe_modes, h, w) complex tensor.
            Should be the psi updated in the reciprocal space step.
        """
        positions = self.forward_model.intermediate_variables["positions"]
        psi_0 = self.forward_model.intermediate_variables["psi"]
        # Shape of chi:           (batch_size, n_probe_modes, h, w)
        chi = psi_opt - psi_0  # Eq, 19
        obj_patches = self.forward_model.intermediate_variables["obj_patches"]

        self.update_reconstruction_parameters(indices, chi, obj_patches, positions)

    @timer()
    def update_reconstruction_parameters(self, indices, chi, obj_patches, positions):
        """
        Update the object and probe.

        Parameters
        ----------
        indices : Tensor.
            A tensor of indices of diffraction patterns processed in the current batch.
        chi : Tensor.
            A (batch_size, n_modes, h, w) complex tensor giving the exit wave difference
            (psi_opt - psi_0) at the exit plane.
        obj_patches : Tensor.
            A (batch_size, h, w) complex tensor giving the object patches.
        positions : Tensor.
            A (batch_size, 2) tensor giving the probe positions in the current batch.
        gamma : float
            Damping factor for solving the step size linear equations.
        """
        object_ = self.parameter_group.object
        self._initialize_object_gradient()
        self._initialize_probe_gradient()
        self._initialize_object_step_size_buffer()
        self._initialize_probe_step_size_buffer()
        self._initialize_momentum_buffers()

        for i_slice in range(object_.n_slices - 1, -1, -1):
            if i_slice < object_.n_slices - 1:
                chi = self.forward_model.propagate_to_previous_slice(chi, slice_index=i_slice + 1)

            # Get unique probes, or the wavefield at the current slice before modulation.
            probe_current_slice = self._get_incident_wavefields_for_slice(i_slice)

            # Caluclate various object-related update directions.
            delta_o_comb, delta_o_precond, delta_o_i, delta_o_i_mode_0 = (
                self.calculate_object_update_directions(
                    chi, probe_current_slice, positions, i_slice
                )
            )

            # Calculate probe update direction.
            delta_p_i_unshifted = self._calculate_probe_update_direction(
                chi, obj_patches=obj_patches, slice_index=i_slice, probe_mode_index=None
            )  # Eq. 24a
            delta_p_i = self.adjoint_shift_probe_update_direction(
                indices, delta_p_i_unshifted, first_mode_only=True
            )
            delta_p_hat = self._precondition_probe_update_direction(delta_p_i)  # Eq. 25a

            # Update OPR modes and weights.
            if i_slice == 0:
                if self.parameter_group.opr_mode_weights.optimization_enabled(self.current_epoch):
                    self.parameter_group.opr_mode_weights.update_variable_probe(
                        self.parameter_group.probe,
                        indices,
                        chi,
                        delta_p_i,
                        delta_p_hat,
                        obj_patches,
                        self.current_epoch,
                        probe_mode_index=0,
                    )

            # Update buffered data for momentum acceleration.
            if i_slice == 0:
                self._update_momentum_buffers(delta_p_hat)

            # Calculate optimal step sizes.
            self.calculate_optimal_step_sizes(
                indices,
                chi,
                obj_patches,
                delta_o_i,
                delta_p_hat,
                delta_o_i_mode_0,
                probe=probe_current_slice,
                slice_index=i_slice,
            )

            # In compact batching mode, object is updated at the end of an epoch using gradients
            # accumulated over all minibatches.
            if self.options.batching_mode in [
                enums.BatchingModes.RANDOM,
                enums.BatchingModes.UNIFORM,
            ]:
                self._record_object_slice_gradient(i_slice, delta_o_precond, add_to_existing=False)
            else:
                self._record_object_slice_gradient(i_slice, delta_o_comb, add_to_existing=False)

            if self.parameter_group.probe_positions.optimization_enabled(
                self.current_epoch
            ) and i_slice == self.parameter_group.probe_positions.get_slice_for_correction(
                object_.n_slices
            ):
                mean_alpha_o = pmath.trim_mean(
                    self.alpha_object_all_pos_all_slices[indices, i_slice], 0.1, dim=0
                )
                self.update_probe_positions(
                    chi,
                    indices,
                    obj_patches[:, i_slice : i_slice + 1],
                    mean_alpha_o * delta_o_i,
                    self.forward_model.intermediate_variables.shifted_unique_probes[i_slice],
                )

            # Set chi to conjugate-modulated wavefield.
            chi = delta_p_i_unshifted

        mean_alpha_o_all_slices = pmath.trim_mean(
            self.alpha_object_all_pos_all_slices[indices], 0.1, dim=0
        )
        if self.parameter_group.object.optimization_enabled(
            self.current_epoch
        ) and self.options.batching_mode in [
            enums.BatchingModes.RANDOM,
            enums.BatchingModes.UNIFORM,
        ]:
            self._apply_object_update(mean_alpha_o_all_slices, None)

    @timer()
    def _get_incident_wavefields_for_slice(self, i_slice):
        r"""
        Get the incident wavefields for a specified slice. For the first slice,
        this is just the unique probes. For other slices in a multislice object,
        this is the wavefield at the given slice before modulation.

        Parameters
        ----------
        i_slice : int
            The current slice index.
        indices : Tensor
            A tensor of indices of diffraction patterns processed in the current batch.

        Returns
        -------
        Tensor
            A (batch_size, n_modes, h, w) tensor giving the incident wavefields.
        """
        incident_wavefields = self.forward_model.intermediate_variables.shifted_unique_probes[
            i_slice
        ]
        return incident_wavefields

    @timer()
    def calculate_object_update_directions(self, chi, probe, positions, i_slice=0):
        """
        Calculate various update directions related to the object. These include
        the preconditioned patch-wise and combined object update directions,
        calculated with the first and all probe modes, respectively.

        Parameters
        ----------
        chi : Tensor
            A (batch_size, n_probe_modes, h, w) tensor giving the exit wave update.
        probe : Tensor
            A (batch_size, n_probe_modes, h, w) tensor giving the unique probe for
            each point.
        positions : Tensor
            A (batch_size, 2) tensor giving the probe positions.
        i_slice : int, optional
            The object slice index.

        Returns
        -------
        delta_o_comb : Tensor
            A (1, h, w) tensor giving the combined object update direction, but not
            yet preconditioned.
        delta_o_precond : Tensor
            A (1, h, w) tensor giving the combined and preconditioned object update
            direction.
        delta_o_i : Tensor
            A (batch_size, 1, h, w) tensor giving the object patch update directions
            after preconditioning.
        delta_o_i_mode_0 : Tensor, optional
            A (batch_size, 1, h, w) tensor giving the object patch update directions
            for the first probe mode after preconditioning (used for step size
            calculation).
        """
        if (
            self.options.solve_step_sizes_only_using_first_probe_mode
            or not self.parameter_group.object.options.multimodal_update
        ):
            # If object step size is to be solved with only the first probe mode,
            # then the delta_o_i used should also be calculated using only the first
            # probe mode.
            delta_o_i_mode_0_raw = self._calculate_object_patch_update_direction(
                chi, incident_wavefields=probe, probe_mode_index=0
            )
            delta_o_comb_mode_0 = self._combine_object_patch_update_directions(
                delta_o_i_mode_0_raw, positions, onto_accumulated=True, slice_index=i_slice
            )
            _, delta_o_i_mode_0 = self._precondition_object_update_direction(
                delta_o_comb_mode_0, 
                positions, 
                alpha_mix=self.options.preconditioning_damping_factor
            )
        else:
            delta_o_i_mode_0 = None

        # Calculate object update direction and precondition it.
        if self.parameter_group.object.options.multimodal_update:
            delta_o_i_raw = self._calculate_object_patch_update_direction(
                chi, incident_wavefields=probe, probe_mode_index=None
            )
            delta_o_comb = self._combine_object_patch_update_directions(
                delta_o_i_raw, positions, onto_accumulated=True, slice_index=i_slice
            )
        else:
            delta_o_i_raw = delta_o_i_mode_0_raw
            delta_o_comb = delta_o_comb_mode_0
        delta_o_precond, delta_o_i = self._precondition_object_update_direction(
            delta_o_comb, 
            positions, 
            alpha_mix=self.options.preconditioning_damping_factor
        )
        return delta_o_comb, delta_o_precond, delta_o_i, delta_o_i_mode_0

    @timer()
    def calculate_optimal_step_sizes(
        self,
        indices,
        chi,
        obj_patches,
        delta_o_i,
        delta_p_hat,
        delta_o_i_mode_0,
        probe=None,
        slice_index=0,
    ):
        """Calculate the optimal step sizes for the object and probem, and update
        the step size buffers.

        Parameters
        ----------
        indices : Tensor
            A tensor of indices of diffraction patterns processed in the current batch.
        chi : Tensor
            A (batch_size, n_probe_modes, h, w) tensor giving the exit wave difference.
        obj_patches : Tensor
            A (batch_size, h, w) tensor giving the object patches.
        delta_o_i : Tensor
            A (batch_size, 1, h, w) tensor giving the object patch update directions,
            calculated using all probe modes.
        delta_p_hat : Tensor
            A (n_probe_modes, h, w) tensor giving the preconditioned probe update direction.
        delta_o_i_mode_0 : Tensor, optional
            A (batch_size, 1, h, w) tensor giving the object patch update directions,
            calculated using the first probe mode.
        probe : Tensor, optional
            A (batch_size, n_probe_modes, h, w) tensor giving the probe.
        slice_index : int, optional
            The slice index of the object.
        """
        object_ = self.parameter_group.object
        if (not object_.is_multislice) or (
            object_.is_multislice
            and slice_index == 0
            and self.options.solve_obj_prb_step_size_jointly_for_first_slice_in_multislice
        ):
            (alpha_o_i, alpha_p_i) = self.calculate_object_and_probe_update_step_sizes(
                chi,
                obj_patches,
                (
                    delta_o_i_mode_0
                    if self.options.solve_step_sizes_only_using_first_probe_mode
                    else delta_o_i
                ),
                delta_p_hat,
                probe=probe,
                probe_mode_index=(
                    0 if self.options.solve_step_sizes_only_using_first_probe_mode else None
                ),
            )
        else:
            alpha_o_i = self.calculate_object_update_step_sizes(
                chi,
                (
                    delta_o_i_mode_0
                    if self.options.solve_step_sizes_only_using_first_probe_mode
                    else delta_o_i
                ),
                probe=probe,
                probe_mode_index=(
                    0 if self.options.solve_step_sizes_only_using_first_probe_mode else None
                ),
            )
            alpha_p_i = self.calculate_probe_update_step_sizes(
                chi,
                obj_patches,
                delta_p_hat,
                probe_mode_index=(
                    0 if self.options.solve_step_sizes_only_using_first_probe_mode else None
                ),
            )

        if slice_index == 0 and self.parameter_group.probe.optimization_enabled(self.current_epoch):
            self._apply_probe_update(alpha_p_i, delta_p_hat)

        self.alpha_object_all_pos_all_slices[indices, slice_index] = alpha_o_i
        self.alpha_probe_all_pos[indices] = alpha_p_i

    @timer()
    def calculate_object_and_probe_update_step_sizes(
        self,
        chi,
        obj_patches,
        delta_o_i,
        delta_p_hat,
        probe=None,
        slice_index=0,
        probe_mode_index=None,
    ):
        """
        Jointly calculate the update step sizes for object and probe according to Eq. 22 of Odstrcil (2018).
        This routine builds a (batch_size, 2, 2) batch matrix, batch-invert them to get the update step sizes.
        """
        mode_slicer = self.parameter_group.probe._get_probe_mode_slicer(probe_mode_index)

        obj_patches = obj_patches[:, slice_index]
        delta_o_i = delta_o_i[:, 0]

        if probe is None:
            probe = self.forward_model.intermediate_variables.shifted_unique_probes[0]
        # When no OPR mode is present, probe is (n_modes, h, w). We add a batch dimension here.
        if probe.ndim == 3:
            probe = probe[None, ...]

        probe = probe[:, mode_slicer]
        chi = chi[:, mode_slicer]
        # TODO: consolidate
        delta_p_hat = delta_p_hat[None, ...]
        if delta_p_hat.shape[1] > 1:
            delta_p_hat = delta_p_hat[:, mode_slicer]

        lambda_0 = 1.2e-7 / (probe.shape[-2] * probe.shape[-1])
        lambda_lsq = 0.1

        # Shape of delta_p_o/o_p:     (batch_size, n_probe_modes or 1, h, w)
        delta_p_o = delta_p_hat * obj_patches[:, None, :, :]
        delta_o_patches_p = delta_o_i[:, None, :, :] * probe

        # Shape of aij:               (batch_size,)
        a11 = torch.sum((delta_o_patches_p.abs() ** 2 + lambda_0), dim=(-1, -2, -3))
        a11 = a11 + lambda_lsq * torch.mean(a11, dim=0)
        a12 = torch.sum((delta_o_patches_p * delta_p_o.conj()), dim=(-1, -2, -3))
        a21 = a12.conj()
        a22 = torch.sum((delta_p_o.abs() ** 2 + lambda_0), dim=(-1, -2, -3))
        a22 = a22 + lambda_lsq * torch.mean(a22, dim=0)
        b1 = torch.sum(torch.real(delta_o_patches_p.conj() * chi), dim=(-1, -2, -3))
        b2 = torch.sum(torch.real(delta_p_o.conj() * chi), dim=(-1, -2, -3))

        a_mat = torch.stack([a11, a12, a21, a22], dim=1).view(-1, 2, 2)
        b_vec = torch.stack([b1, b2], dim=1).view(-1, 2).type(a_mat.dtype)
        alpha_vec = torch.linalg.solve(a_mat, b_vec)
        alpha_vec = alpha_vec.real.clip(0, None)

        alpha_o_i = alpha_vec[:, 0]
        alpha_p_i = alpha_vec[:, 1]

        alpha_o_i = alpha_o_i * self.parameter_group.object.options.optimal_step_size_scaler
        alpha_p_i = alpha_p_i * self.parameter_group.probe.options.optimal_step_size_scaler

        alpha_o_i = alpha_o_i / self.parameter_group.object.n_slices
        if self.parameter_group.object.options.multimodal_update:
            alpha_o_i = alpha_o_i / self.parameter_group.probe.n_modes

        logger.debug("alpha_p_i: min={}, max={}".format(alpha_p_i.min(), alpha_p_i.max()))
        logger.debug("alpha_o_i: min={}, max={}".format(alpha_o_i.min(), alpha_o_i.max()))

        return alpha_o_i, alpha_p_i

    @timer()
    def calculate_object_update_step_sizes(self, chi, delta_o_i, probe=None, probe_mode_index=None):
        """
        Calculate the update step sizes just for the object using Eq. 23b of Odstrcil (2018).
        """
        # Just take the first slice.
        delta_o_i = delta_o_i[:, 0]

        mode_slicer = self.parameter_group.probe._get_probe_mode_slicer(probe_mode_index)

        if probe is None:
            probe = self.forward_model.intermediate_variables.shifted_unique_probes[0]

        probe = probe[:, mode_slicer]
        chi = chi[:, mode_slicer]

        # Shape of delta_p_o/o_p:     (batch_size, n_probe_modes or 1, h, w)
        delta_o_patches_p = delta_o_i[:, None, :, :] * probe

        numerator = 0.5 * torch.sum(torch.real(delta_o_patches_p.conj() * chi), dim=(-1, -2, -3))
        denominator = torch.sum(delta_o_patches_p.abs() ** 2, dim=(-1, -2, -3))

        alpha_o_i = numerator / denominator
        alpha_o_i = alpha_o_i * self.parameter_group.object.options.optimal_step_size_scaler
        alpha_o_i = alpha_o_i / self.parameter_group.object.n_slices
        if self.parameter_group.object.options.multimodal_update:
            alpha_o_i = alpha_o_i / self.parameter_group.probe.n_modes

        alpha_o_i = alpha_o_i.clamp(0, None)

        logger.debug(
            "alpha_o_i: min={}, max={}, trim_mean={}".format(
                alpha_o_i.min(), alpha_o_i.max(), pmath.trim_mean(alpha_o_i)
            )
        )
        return alpha_o_i

    @timer()
    def calculate_probe_update_step_sizes(
        self, chi, obj_patches, delta_p_hat, probe_mode_index=None
    ):
        """
        Calculate the update step sizes just for the probe using Eq. 23a of Odstrcil (2018).
        """
        # Just take the first slice.
        obj_patches = obj_patches[:, 0]

        mode_slicer = self.parameter_group.probe._get_probe_mode_slicer(probe_mode_index)

        delta_p_hat = delta_p_hat[None, mode_slicer]
        chi = chi[:, mode_slicer]

        # Shape of delta_p_o/o_p:     (batch_size, n_probe_modes, h, w)
        delta_p_o = delta_p_hat * obj_patches[:, None, :, :]

        # Shape of aij:               (batch_size,)
        numerator = 0.5 * torch.sum(torch.real(delta_p_o.conj() * chi), dim=(-1, -2, -3))
        denominator = torch.sum(delta_p_o.abs() ** 2, dim=(-1, -2, -3))

        alpha_p_i = numerator / denominator
        alpha_p_i = alpha_p_i * self.parameter_group.probe.options.optimal_step_size_scaler

        alpha_p_i = alpha_p_i.clamp(0, None)

        logger.debug("alpha_p_i: min={}, max={}".format(alpha_p_i.min(), alpha_p_i.max()))
        return alpha_p_i

    @timer()
    def _calculate_probe_update_direction(
        self, chi, obj_patches=None, slice_index=0, probe_mode_index=None
    ):
        """
        Calculate probe update direction using Eq. 24a of Odstrcil (2018).

        Parameters
        ----------
        chi: torch.Tensor
            A (batch_size, n_probe_modes, h, w) tensor giving the difference of exit waves.
        obj_patches: torch.Tensor
            A (batch_size, h, w) tensor giving the object patches. If None, just return
            chi as it is. This behavior is intended for multislice.
        slice: int
            The slice of the object patches used to calculate the update direction.
        """
        mode_slicer = self.parameter_group.probe._get_probe_mode_slicer(probe_mode_index)

        if obj_patches is not None:
            obj_patches = obj_patches[:, slice_index]
            delta_p = chi[:, mode_slicer] * obj_patches.conj()[:, None, :, :]  # Eq. 24a
        else:
            delta_p = chi[:, mode_slicer]
        return delta_p

    @timer()
    def _precondition_probe_update_direction(self, delta_p):
        """
        Eq. 25a of Odstrcil, 2018.

        Parameters
        ----------
        delta_p : Tensor
            A (batch_size, n_probe_modes, h, w) tensor giving the probe update direction.

        Returns
        -------
        Tensor
            A (n_probe_modes, h, w) tensor giving the preconditioned probe update direction.
        """
        # Shape of delta_p_hat:  (n_probe_modes, h, w)
        delta_p_hat = torch.sum(delta_p, dim=0)  # Eq. 25a
        # PtychoShelves code simply takes the average. This is different from the paper
        # which does delta_p_hat = delta_p_hat / ((object_.abs() ** 2).sum() + delta),
        # but this seems to work better.
        delta_p_hat = delta_p_hat / delta_p.shape[0]
        return delta_p_hat

    @timer()
    def _apply_probe_update(self, alpha_p_i, delta_p_hat, probe_mode_index=None):
        """
        Eq. 27a of Odstrcil, 2018.
        """
        # Shape of alpha_p_i:        (batch_size,)
        # Shape of delta_p_hat:      (n_probe_modes, h, w)
        # PtychoShelves code simply multiplies delta_p_hat with averaged step size.
        # This is different from the paper which does the following:
        #     update_vec = delta_p_hat * obj_patches[:, None, :, :].abs() ** 2
        #     update_vec = update_vec * alpha_p_i[:, None, None, None]
        #     update_vec = update_vec / ((obj_patches.abs() ** 2).sum(0) + delta)

        # Just apply the update to the main OPR mode of each incoherent mode.
        # To do this, we pad the update vector with zeros in the OPR mode dimension.
        mode_slicer = self.parameter_group.probe._get_probe_mode_slicer(probe_mode_index)

        if self.options.batching_mode == enums.BatchingModes.COMPACT:
            # In compact mode, object is updated only once per epoch. To match the probe to this,
            # we divide the probe step size by the number of minibatches before each probe update.
            alpha_p_i = alpha_p_i / len(self.dataloader)
        alpha_p_mean = torch.mean(alpha_p_i)
        self.parameter_group.probe.set_grad(-delta_p_hat * alpha_p_mean, slicer=(0, mode_slicer))
        self.parameter_group.probe.optimizer.step()

    @timer()
    def _apply_probe_momentum(self, alpha_p_mean, delta_p_hat):
        """
        Apply momentum acceleration to the probe (only the first OPR mode). This is a
        special momentum acceleration used in PtychoShelves, which behaves somewhat
        differently from the momentum in `torch.optim.SGD`.

        Parameters
        ----------
        alpha_p_mean: float
            A scalar giving the mean probe step size.
        delta_p_hat: torch.Tensor
            A (n_probe_modes, h, w) tensor giving the accumulated probe update direction
            of the first OPR mode.
        """
        delta_p_hat = delta_p_hat * alpha_p_mean

        probe = self.parameter_group.probe
        if "update_direction_history" not in self.probe_momentum_params.keys():
            self.probe_momentum_params["update_direction_history"] = []
            self.probe_momentum_params["velocity_map"] = torch.zeros_like(delta_p_hat)

        upd = delta_p_hat / (pmath.mnorm(delta_p_hat, dim=(-1, -2), keepdims=True) + 1e-15)
        self.probe_momentum_params["update_direction_history"].append(upd)

        momentum_memory = 3

        if len(self.probe_momentum_params["update_direction_history"]) > momentum_memory + 1:
            # Remove the oldest momentum update.
            self.probe_momentum_params["update_direction_history"].pop(0)
            # PtychoShelves only applies momentum to the first mode.
            for i_mode in range(1):
                # Project older updates to the latest one, ordered from recent to old.
                projected_updates = [
                    (
                        self.probe_momentum_params["update_direction_history"][i][i_mode]
                        * self.probe_momentum_params["update_direction_history"][-1][i_mode].conj()
                    )
                    for i in range(len(self.probe_momentum_params["update_direction_history"]) - 1)
                ][::-1]
                # Shape of corr_level: (momentum_memory, n_probe_modes)
                corr_level = [
                    torch.mean(projected_updates[k]).real for k in range(len(projected_updates))
                ]
                corr_level = torch.tensor(corr_level, device=delta_p_hat.device)

                if self._fourier_error_ok() and torch.all(corr_level > 0):
                    # Estimate optimal friction.
                    p = pmath.polyfit(
                        torch.arange(0.0, momentum_memory + 1),
                        torch.concat([torch.zeros([1]), torch.log(corr_level)], dim=0).reshape(-1),
                        deg=1,
                    )
                    friction = 0.5 * (-p[0]).clip(0, None)

                    m = self.options.momentum_acceleration_gradient_mixing_factor
                    m = friction if m is None else m
                    self.probe_momentum_params["velocity_map"][i_mode] = (
                        1 - friction
                    ) * self.probe_momentum_params["velocity_map"][i_mode] + m * delta_p_hat[i_mode]
                    probe.set_data(
                        probe.data[0, i_mode]
                        + self.options.momentum_acceleration_gain
                        * self.probe_momentum_params["velocity_map"][i_mode],
                        slicer=(0, i_mode),
                    )
                else:
                    self.probe_momentum_params["velocity_map"][i_mode] = (
                        self.probe_momentum_params["velocity_map"][i_mode] / 2.0
                    )

    @timer()
    def _calculate_object_patch_update_direction(
        self, chi, incident_wavefields=None, probe_mode_index=None
    ):
        r"""
        Calculate the update direction for object patches, implementing
        Eq. 24b of Odstrcil, 2018. This function works in both 2D mode and
        multislice mode:

        - When `incident_wavefields` is None, 2D mode is assumed. `chi` is multiplied with the
            complex conjugate of the probe.
        - When `incident_wavefields` is not None, multislice mode is assumed. `chi` is multiplied
            with the complex conjugate of `incident_wavefields`.

        Parameters
        ----------
        indices : Tensor
            Indices of diffraction patterns in the current batch.
        chi : Tensor
            A (batch_size, n_modes, h, w) tensor giving the difference of exit waves.
            For multislice, this should be the exiting-plane `chi` backpropagated to the
            current slice.
        incident_wavefields : Tensor
            A (batch_size, n_modes, h, w) tensor giving $\psi_{i - 1}$, the wavefield
            modulated by the previous slice and propagated to the current slice. If this
            is given, multislice mode is assumed.

        Returns
        -------
        Tensor
            A (batch_size, 1, h, w) tensor giving the update direction for object patches.
            The dimension of size 1 is to match the slice dimension in the object patch
            tensor.
        """
        mode_slicer = self.parameter_group.probe._get_probe_mode_slicer(probe_mode_index)

        if incident_wavefields is None:
            p = self.forward_model.intermediate_variables.shifted_unique_probes[0]
        else:
            p = incident_wavefields

        if p.ndim == 3:
            p = p[None, ...]

        p = p[:, mode_slicer]
        chi = chi[:, mode_slicer]
        # Shape of chi:          (batch_size, n_probe_modes, h, w)
        # Shape delta_o_patches: (batch_size, h, w)
        # Multiply and sum over probe mode dimension
        delta_o_patches = torch.sum(chi * p.conj(), dim=1)  # Eq. 24b

        # Add slice dimension.
        return delta_o_patches[:, None, :, :]

    @timer()
    def _combine_object_patch_update_directions(
        self, delta_o_patches, positions, onto_accumulated=False, slice_index=0
    ):
        """
        Combine the update directions of object patches into a buffer with the
        same size as the whole object.

        Parameters
        ----------
        delta_o_patches : Tensor
            A (batch_size, 1, h, w) tensor giving the update direction for object patches.
        onto_accumulated : bool
            If True, add the update direction to the accumulated update direction stored in
            `object.grad`. Otherwise, add the update direction to the object buffer.

        Returns
        -------
        Tensor
            A (1, h, w) tensor giving the combined update direction for the whole object.
        """
        delta_o_patches = delta_o_patches[:, 0]

        # Stitch all delta O patches on the object buffer
        # Shape of delta_o_hat:  (h_whole, w_whole)
        delta_o_hat = self.parameter_group.object.place_patches_on_empty_buffer(
            positions.round().int(), delta_o_patches, integer_mode=True
        )
        delta_o_hat = delta_o_hat[None, ...]
        if onto_accumulated:
            delta_o_hat = delta_o_hat + (
                -self.parameter_group.object.get_grad()[slice_index : slice_index + 1]
            )
        return delta_o_hat

    @timer()
    def _precondition_object_update_direction(
        self, delta_o_hat, positions=None, alpha_mix=0.1, slice_index=0
    ):
        """
        Eq. 25b of Odstrcil, 2018.

        Returns
        -------
        Tensor
            A (1, h, w) tensor giving the preconditioned update direction for the whole object.
        Tensor
            A (batch_size, 1, h, w) tensor giving the preconditioned update direction for object patches.
            Only returned when `positions` is not None.
        """
        delta_o_hat = delta_o_hat[slice_index]

        preconditioner = self.parameter_group.object.preconditioner
        delta_o_hat = delta_o_hat / torch.sqrt(
            preconditioner**2 + (preconditioner.max() * alpha_mix) ** 2
        )

        # Re-extract delta O patches
        if positions is not None:
            delta_o_patches = ip.extract_patches_integer(
                delta_o_hat,
                positions.round().int() + self.parameter_group.object.pos_origin_coords,
                self.parameter_group.probe.shape[-2:],
            )

            return delta_o_hat[None, ...], delta_o_patches[:, None, :, :]
        return delta_o_hat[None, ...]

    @timer()
    def _precondition_accumulated_object_update_direction(self):
        """
        Sequentially precondition the object update direction accumulated over minibatches
        and stored in `object.grad`. This is only used in compact batching mode, where the
        object is only updated at the end of each epoch.
        """
        delta_o_hat_full = []
        for i_slice in range(self.parameter_group.object.n_slices):
            delta_o_hat = self._precondition_object_update_direction(
                -self.parameter_group.object.get_grad()[i_slice : i_slice + 1], 
                positions=None,
                alpha_mix=self.options.preconditioning_damping_factor
            )
            delta_o_hat_full.append(delta_o_hat)
        delta_o_hat_full = torch.cat(delta_o_hat_full, dim=0)
        return delta_o_hat_full

    @timer()
    def _initialize_object_gradient(self):
        """
        Initialize object gradient with zeros. This method is called at the beginning of the
        real-space step of a minibatch. If batching mode is "random/uniform", the gradient is always
        re-initialized when this method is called. If batching mode is "compact", gradient
        is only initialized if the current minibatch is the first in the current epoch.
        """
        if self.options.batching_mode in [enums.BatchingModes.RANDOM, enums.BatchingModes.UNIFORM]:
            self.parameter_group.object.initialize_grad()
        else:
            if self.current_minibatch == 0:
                self.parameter_group.object.initialize_grad()

    @timer()
    def _initialize_probe_gradient(self):
        self.parameter_group.probe.initialize_grad()

    @timer()
    def _initialize_object_step_size_buffer(self):
        if self.current_minibatch == 0:
            self.alpha_object_all_pos_all_slices[...] = 1

    @timer()
    def _initialize_probe_step_size_buffer(self):
        if self.current_minibatch == 0:
            self.alpha_probe_all_pos[...] = 1

    @timer()
    def _initialize_momentum_buffers(self):
        """Initialize momentum buffers.

        Only the probe's accumulated update direction is initialized here. The accumulated update
        direction for the object is stored in `object.grad`.
        """
        if self.options.batching_mode != enums.BatchingModes.COMPACT or self.current_minibatch == 0:
            self.probe_momentum_params["accumulated_update_direction"] = 0

    @timer()
    def _update_momentum_buffers(self, delta_p_hat):
        """Update momentum buffer for probe after each minibatch using the update direction calculated
        in that minibatch.

        We do not track the update direction for the object here, because it is already recorded in
        `object.grad`.

        Parameters
        ----------
        delta_p_hat : Tensor
            A (n_opr_modes, n_probe_modes, h, w) tensor giving the update direction for the probe.
        """
        self.probe_momentum_params["accumulated_update_direction"] += delta_p_hat / len(
            self.dataloader
        )

    @timer()
    def _record_object_slice_gradient(self, i_slice, delta_o_hat, add_to_existing=False):
        """
        Record the gradient of one slice of a multislice object.
        """
        if not add_to_existing:
            self.parameter_group.object.set_grad(-delta_o_hat[0], slicer=i_slice)
        else:
            self.parameter_group.object.set_grad(
                self.parameter_group.object.get_grad()[i_slice] - delta_o_hat[0], slicer=i_slice
            )

    @timer()
    def _apply_object_update(self, alpha_o_mean_all_slices, delta_o_hat=None):
        """
        Apply object update using Eq. 27b of Odstrcil, 2018.

        If both `alpha_o_mean` and `delta_o_hat` are given, the object's gradient is set
        using averaged step size and `delta_o_hat`. Otherwise, we assume the gradient
        is already set previously, and just simply run the optimizer step.

        Parameters
        ----------
        alpha_o_mean_all_slices : Tensor
            A (n_slices,) tensor giving the averaged step size for each object slice.
        delta_o_hat : Tensor
            A (n_slices, h, w) tensor giving the update direction for the whole object. If None,
            use the (negative) update direction stored in `object.grad`.
        """
        if delta_o_hat is None:
            delta_o_hat = -self.parameter_group.object.get_grad()
        self.parameter_group.object.set_grad(-alpha_o_mean_all_slices[:, None, None] * delta_o_hat)
        self.parameter_group.object.optimizer.step()

    @timer()
    def _apply_object_momentum(self, alpha_o_mean_all_slices, delta_o_hat):
        """
        Apply momentum acceleration to the object. This is a special momentum acceleration used
        in PtychoShelves, which behaves somewhat differently from the momentum in `torch.optim.SGD`.
        """
        # Scale object update by step size at the beginning to match the behavior in PtychoShelves,
        # In PtychoShelves, this scaling happens in `update_object.m`.
        delta_o_hat = delta_o_hat * alpha_o_mean_all_slices[:, None, None]

        object_ = self.parameter_group.object
        if "update_direction_history" not in self.object_momentum_params.keys():
            self.object_momentum_params["update_direction_history"] = []
            self.object_momentum_params["velocity_map"] = torch.zeros_like(object_.data)

        object_roi_bbox = self.parameter_group.object.roi_bbox.get_bbox_with_top_left_origin()
        object_roi_slicer = object_roi_bbox.get_slicer()
        upd = delta_o_hat * alpha_o_mean_all_slices[:, None, None]
        upd = upd[:, *object_roi_slicer]
        upd = upd / pmath.mnorm(upd, dim=(-1, -2), keepdims=True)
        self.object_momentum_params["update_direction_history"].append(upd)

        momentum_memory = 2

        if len(self.object_momentum_params["update_direction_history"]) > momentum_memory + 1:
            # Remove the oldest momentum update.
            self.object_momentum_params["update_direction_history"].pop(0)
            for i_slice in range(object_.n_slices):
                if self._fourier_error_ok():
                    # Project older updates to the latest one, ordered from recent to old.
                    projected_updates = [
                        (
                            self.object_momentum_params["update_direction_history"][i][i_slice]
                            * self.object_momentum_params["update_direction_history"][-1][
                                i_slice
                            ].conj()
                        )
                        for i in range(
                            len(self.object_momentum_params["update_direction_history"]) - 1
                        )
                    ][::-1]
                    corr_level = [
                        torch.mean(projected_updates[k]).real for k in range(len(projected_updates))
                    ]
                    corr_level = torch.tensor(corr_level, device=delta_o_hat.device)

                if self._fourier_error_ok() and torch.all(corr_level > 0):
                    # Estimate optimal friction.
                    p = pmath.polyfit(
                        torch.arange(0.0, momentum_memory + 1.0, device=delta_o_hat.device),
                        torch.tensor([0, *torch.log(corr_level)], device=delta_o_hat.device),
                        deg=1,
                    )
                    # If correlation drops fast as one goes away from the current epoch, p[0]
                    # is more negative, and friction is larger. Accumulated velocity is weighted
                    # less to allow the update direction to change more quickly.
                    friction = 0.5 * (-p[0]).clip(0, None)

                    w = object_.preconditioner / (
                        0.1 * object_.preconditioner.max() + object_.preconditioner
                    )
                    m = self.options.momentum_acceleration_gradient_mixing_factor
                    m = friction if m is None else m
                    self.object_momentum_params["velocity_map"][i_slice] = (
                        1 - friction
                    ) * self.object_momentum_params["velocity_map"][i_slice] + m * delta_o_hat[
                        i_slice
                    ]
                    object_.set_data(
                        object_.data[i_slice]
                        + w
                        * self.options.momentum_acceleration_gain
                        * self.object_momentum_params["velocity_map"][i_slice],
                        slicer=i_slice,
                    )
                else:
                    self.object_momentum_params["velocity_map"][i_slice] = (
                        self.object_momentum_params["velocity_map"][i_slice] / 2.0
                    )

    @timer()
    def _fourier_error_ok(self):
        if len(self.fourier_errors) < 3:
            return True
        return max(self.fourier_errors[-3:-1]) > min(self.fourier_errors[-2:])

    @timer()
    def update_probe_positions(self, chi, indices, obj_patches, delta_o_patches, unique_probes):
        delta_pos = self.parameter_group.probe_positions.position_correction.get_update(
            chi,
            obj_patches,
            delta_o_patches,
            unique_probes,
            self.parameter_group.object.optimizer_params["lr"],
        )
        self._apply_probe_position_update(delta_pos, indices)

    @timer()
    def _apply_probe_position_update(self, delta_pos, indices):
        delta_pos_full = torch.zeros_like(self.parameter_group.probe_positions.tensor)
        delta_pos_full[indices] = delta_pos
        self.parameter_group.probe_positions.set_grad(-delta_pos_full)
        self.parameter_group.probe_positions.step_optimizer()

    @timer()
    def _calculate_final_object_update_step_size(self):
        """
        Given the patch-wise step sizes, calculate the final step size for the whole object used
        in compact-mode update.

        This routine follows the same logic as in PtychoSheleves. With the `(n_pos, n_slices)`
        tensor that sotres the step sizes for all object patches and all slices, we take the
        10-th percentile trimmed mean of the step sizes for each minibatch. We then take the
        minimum of the step sizes across all minibatches for each slice to use as the step size
        for updating the object.

        Returns
        -------
        Tensor
            A (n_slices,) tensor giving the final step size for each slice.
        """
        alpha_object_all_minibatches = []
        for inds in self.indices:
            alpha_current_batch = pmath.trim_mean(
                self.alpha_object_all_pos_all_slices[inds], 0.1, dim=0, keepdim=False
            )
            alpha_object_all_minibatches.append(alpha_current_batch)
        alpha_object_all_minibatches = torch.stack(alpha_object_all_minibatches, dim=0)
        alpha_object_all_slices = torch.min(
            alpha_object_all_minibatches, dim=0, keepdim=False
        ).values
        return alpha_object_all_slices

    @timer()
    def _update_accumulated_intensities(self, y_true, y_pred):
        self.accumulated_true_intensity = self.accumulated_true_intensity + torch.sum(y_true)
        self.accumulated_pred_intensity = self.accumulated_pred_intensity + torch.sum(y_pred)

    @timer()
    def _apply_probe_intensity_scaling_correction(self):
        corr = math.sqrt(self.accumulated_true_intensity / self.accumulated_pred_intensity)
        self.parameter_group.probe.set_data(self.parameter_group.probe.data * corr)

    @timer()
    def update_fourier_error(self, y_pred, y_true):
        self.accumulated_fourier_error += torch.mean(
            (torch.sqrt(y_pred) - torch.sqrt(y_true)) ** 2, dim=(-2, -1)
        ).sum()
        if self.current_minibatch == len(self.dataloader) - 1:
            e = self.accumulated_fourier_error / self.parameter_group.probe_positions.shape[0]
            self.fourier_errors.append(e.item())
            self.accumulated_fourier_error = 0.0

    def run_pre_run_hooks(self) -> None:
        self.prepare_data()

    def run_pre_epoch_hooks(self) -> None:
        super().run_pre_epoch_hooks()
        self.accumulated_fourier_error = 0.0
        self.indices = []

    @timer()
    def run_post_epoch_hooks(self) -> None:
        if self.current_epoch > 0 or (not self.options.rescale_probe_intensity_in_first_epoch):
            if (
                self.parameter_group.object.optimization_enabled(self.current_epoch)
                and self.options.batching_mode == enums.BatchingModes.COMPACT
            ):
                # Take the 10-th percentile of the object step sizes across all minibatches for
                # each slice to use as the step size for updating the object.
                alpha_object_all_slices = self._calculate_final_object_update_step_size()
                delta_o_hat_full = self._precondition_accumulated_object_update_direction()
                self._apply_object_update(alpha_object_all_slices, delta_o_hat_full)

                if self.options.momentum_acceleration_gain > 0:
                    # Momentum acceleration for object is only applied for compact batching.
                    self._apply_object_momentum(alpha_object_all_slices, delta_o_hat_full)

            if (
                self.parameter_group.probe.optimization_enabled(self.current_epoch)
                and self.options.momentum_acceleration_gain > 0
            ):
                self._apply_probe_momentum(
                    torch.mean(self.alpha_probe_all_pos),
                    self.probe_momentum_params["accumulated_update_direction"],
                )
        else:
            # In epoch 0, only correct probe intensity.
            self._apply_probe_intensity_scaling_correction()
        return super().run_post_epoch_hooks()

    @timer()
    def run_minibatch(self, input_data, y_true, *args, **kwargs) -> None:
        indices = input_data[0]
        self.indices.append(indices)
        y_pred = self.forward_model(*input_data)
        self.update_fourier_error(y_pred, y_true)
        if self.current_epoch == 0 and self.options.rescale_probe_intensity_in_first_epoch:
            self._update_accumulated_intensities(y_true, y_pred)
        else:
            psi_opt = self.run_reciprocal_space_step(y_pred, y_true, indices)
            self.run_real_space_step(psi_opt, indices)

        self.loss_tracker.update_batch_loss_with_metric_function(y_pred, y_true)

    def get_config_dict(self) -> dict:
        d = super().get_config_dict()
        d.update({"noise_model": self.noise_model.noise_statistics})
        return d



import pytest
import os
from ptychi.timing.io import get_timestamp_for_timing_files


def pytest_addoption(parser):
    parser.addoption(
        "--high-tol", action="store_true", help="Use high tolerance for certain tests."
    )
    parser.addoption("--all", action="store_true", help="Run all tests.")
    parser.addoption("--local-only", action="store_true", help="Just run local-only test cases.")
    parser.addoption(
        "--action",
        default="all",
        help="Select the action to run. Can be: "
             "\"generate\" - just generate outputs and store them; "
             "\"compare\" - just run comparison with reference data; "
             "\"all\" - run everything.",
    )
    parser.addoption(
        "--save-timing", action="store_true", help="Enable function timing and save results."
    )


def pytest_configure(config):
    config.addinivalue_line("markers", "local: mark test as local-only")
    # Save timestamps for timing file metadata
    timestamp, date_string, time_string = get_timestamp_for_timing_files()
    os.environ["TIMESTAMP"] = timestamp
    os.environ["DATE_STRING"] = date_string
    os.environ["TIME_STRING"] = time_string


def pytest_collection_modifyitems(config, items):
    if config.getoption("--all"):
        return
    skip_local = pytest.mark.skip(reason="need --all option to run")
    skip_non_local = pytest.mark.skip(reason="skipped due to --local-only flag")
    for item in items:
        if "local" in item.keywords and not config.getoption("--local-only"):
            item.add_marker(skip_local)
        elif "local" not in item.keywords and config.getoption("--local-only"):
            item.add_marker(skip_non_local)



# Copyright © 2025 UChicago Argonne, LLC All right reserved
# Full license accessible at https://github.com//AdvancedPhotonSource/pty-chi/blob/main/LICENSE

from typing import Optional, TYPE_CHECKING

import torch
from torch.utils.data import Dataset

from ptychi.reconstructors.base import (
    AnalyticalIterativePtychographyReconstructor,
)
from ptychi.metrics import MSELossOfSqrt
from ptychi.maths import reprod, redot
from ptychi.utils import get_default_complex_dtype
import ptychi.forward_models as fm

if TYPE_CHECKING:
    import ptychi.api as api
    import ptychi.data_structures.parameter_group as pg


class BHReconstructor(AnalyticalIterativePtychographyReconstructor):
    """
    Bilinear Hessian Reconstruction Method:
    Implements Gradient Descent and Conjugate Gradient algorithms (Carlsson & Nikitin, 2025, in prep).

    So far this implementation supports the reconstruction of the object, probe, and position parameters.
    Probe multi modes need to be implemented.

    The algorithm accepts a parameter `rho`, which controls the optimization of the probe and position updates.
    The value of `rho` is adjusted based on data size and the initial guess.
    Higher values of `rho` give more weight to the updates, while lower values reduce their influence on the reconstruction.

    It is important to adjust the `rho` values carefully to avoid converging to local minima.
    """

    parameter_group: "pg.PlanarPtychographyParameterGroup"

    def __init__(
        self,
        parameter_group: "pg.PlanarPtychographyParameterGroup",
        dataset: Dataset,
        options: Optional["api.options.bh.BHReconstructorOptions"] = None,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(
            parameter_group=parameter_group,
            dataset=dataset,
            options=options,
            *args,
            **kwargs,
        )

        # constant to avoid divisions by 0 in calculations
        self.eps = 1e-8
        # cg varibales
        if self.options.method == "CG":
            self.eta_op = torch.empty(dataset.patterns.shape,dtype=get_default_complex_dtype())[:,None,:,:]
            self.eta_pos = torch.empty([dataset.patterns.shape[0],2])
            
    def build_forward_model(self):
        # BHReconstructor requires subpixel shifts on the object, not the probe.
        self.forward_model = fm.PlanarPtychographyForwardModel(
            parameter_group=self.parameter_group, 
            retain_intermediates=True,
            detector_size=tuple(self.dataset.patterns.shape[-2:]),
            wavelength_m=self.dataset.wavelength_m,
            free_space_propagation_distance_m=self.dataset.free_space_propagation_distance_m,
            pad_for_shift=self.options.forward_model_options.pad_for_shift,
            low_memory_mode=self.options.forward_model_options.low_memory_mode,
            apply_subpixel_shifts_on_probe=False
        )

    def build_loss_tracker(self):
        if self.displayed_loss_function is None:
            self.displayed_loss_function = MSELossOfSqrt()
        return super().build_loss_tracker()

    def check_inputs(self, *args, **kwargs):
        if self.parameter_group.object.is_multislice:
            raise NotImplementedError("The BH only supports 2D objects.")
        if self.parameter_group.probe.has_multiple_opr_modes:
            raise NotImplementedError("The BH does not support multiple OPR modes yet.")

    def run_minibatch(self, input_data, y_true, *args, **kwargs):
        (delta_o, delta_p, delta_pos), y_pred = self.compute_updates(*input_data, y_true)
        self.apply_updates(delta_o, delta_p, delta_pos)
        self.loss_tracker.update_batch_loss_with_metric_function(y_pred, y_true)

    def apply_updates(self, delta_o, delta_p, delta_pos, *args, **kwargs):
        """
        Apply updates to optimizable parameters given the updates calculated by self.compute_updates.

        Parameters
        ----------
        delta_o : Tensor
            A (n_replica, h, w, 2) tensor of object update vector.
        delta_p : Tensor
            A (n_replicate, n_opr_modes, n_modes, h, w, 2) tensor of probe update vector.
        delta_pos : Tensor
            A (n_positions, 2) tensor of probe position vectors.
        """
        object_ = self.parameter_group.object
        probe = self.parameter_group.probe
        probe_positions = self.parameter_group.probe_positions

        if delta_o is not None:
            object_.set_grad(-delta_o)
            object_.optimizer.step()

        if delta_p is not None:
            probe.set_grad(-delta_p)
            probe.optimizer.step()

        if delta_pos is not None:
            probe_positions.set_grad(-delta_pos)
            probe_positions.optimizer.step()

    def compute_updates(
        self, indices: torch.Tensor, y_true: torch.Tensor
    ) -> tuple[torch.Tensor, ...]:
        """
        Calculates the updates of the whole object, the probe, and other parameters.
        This function is called in self.update_step_module.forward.
        """
        object_ = self.parameter_group.object
        probe = self.parameter_group.probe
        probe_positions = self.parameter_group.probe_positions

        self.indices = indices.cpu()
        self.positions = probe_positions.tensor[indices]

        y = self.forward_model.forward(indices)
        op = self.forward_model.intermediate_variables["obj_patches"]

        psi_far = self.forward_model.intermediate_variables["psi_far"]
        p = probe.get_opr_mode(0)  # to do for multi modes
        pos = self.positions

        # sqrt of data
        d = torch.sqrt(y_true)[:, torch.newaxis]

        # Gradient for the Gaussian model
        gradF = self.gradientF(psi_far, d)

        # shortcuts
        o_opt = object_.optimization_enabled(self.current_epoch)
        p_opt = probe.optimization_enabled(self.current_epoch)
        pos_opt = probe_positions.optimization_enabled(self.current_epoch)

        # Calculate updates either of 3 problems: (o,p,pos), (o,p), or (o)
        if o_opt and p_opt and pos_opt:
            (delta_o, delta_p, delta_pos0) = self.compute_updates_object_probe_positions(
                op, p, pos, psi_far, gradF, d
            )

            # operations from the original code
            delta_o = delta_o.unsqueeze(0)
            delta_p_all_modes = delta_p[None, :, :]
            delta_pos = torch.zeros_like(probe_positions.data)
            delta_pos[indices] = delta_pos0

        elif o_opt and p_opt and (not pos_opt):
            (delta_o, delta_p) = self.compute_updates_object_probe(op, p, psi_far, gradF, d)
            delta_pos = None

            # operations from the original code
            delta_o = delta_o.unsqueeze(0)
            delta_p_all_modes = delta_p[None, :, :]

        elif o_opt:
            delta_o = self.compute_updates_object(op, p, psi_far, gradF, d)
            delta_p_all_modes = None
            delta_pos = None

            # operations from the original code
            delta_o = delta_o.unsqueeze(0)

        return (delta_o, delta_p_all_modes, delta_pos), y

    def compute_updates_object(self, op, p, psi_far, gradF, d):
        grad_o, grad_op = self.gradient_o(p, gradF)

        if self.current_epoch == 0 or self.options.method == "GD":
            eta_o = -grad_o
            eta_op = -grad_op
        elif self.options.method == "CG":
            eta_o = self.eta_o
            eta_op = self.eta_op[self.indices]

            beta = self.calc_beta_object(p, d, psi_far, grad_op, eta_op)
            eta_o = -grad_o + beta * eta_o
            eta_op = -grad_op + beta * eta_op

        if self.options.method == "CG":
            self.eta_o = eta_o
            self.eta_op[self.indices] = eta_op

        alpha = self.calc_alpha_object(p, grad_o, eta_o, d, psi_far, eta_op)

        # normalize with the batch size
        alpha *= op.shape[0] / self.batch_size

        delta_o = alpha * eta_o

        return delta_o

    def compute_updates_object_probe(self, op, p, psi_far, gradF, d):
        grad_o, grad_op = self.gradient_o(p, gradF)
        grad_p = self.parameter_group.probe.options.rho * self.gradient_p(op, gradF)

        if self.current_epoch == 0 or self.options.method == "GD":
            eta_o = -grad_o
            eta_op = -grad_op
            eta_p = -grad_p
        elif self.options.method == "CG":
            # intermediate variables for the CG
            eta_o = self.eta_o
            eta_op = self.eta_op[self.indices]
            eta_p = self.eta_p

            beta = self.calc_beta_object_probe(
                p, grad_p, eta_p, d, gradF, psi_far, op, grad_op, eta_op
            )
            eta_o = -grad_o + beta * eta_o
            eta_op = -grad_op + beta * eta_op
            eta_p = -grad_p + beta * eta_p

        if self.options.method == "CG":
            self.eta_o = eta_o
            self.eta_op[self.indices] = eta_op
            self.eta_p = eta_p

        alpha = self.calc_alpha_object_probe(
            p, grad_o, grad_p, eta_o, eta_p, d, gradF, psi_far, op, eta_op
        )

        # normalize with the batch size
        alpha *= op.shape[0] / self.batch_size

        delta_o = alpha * eta_o
        delta_p = self.parameter_group.probe.options.rho * alpha * eta_p
        return (delta_o, delta_p)

    def compute_updates_object_probe_positions(self, op, p, pos, psi_far, gradF, d):
        grad_o, grad_op = self.gradient_o(p, gradF)
        grad_p = self.parameter_group.probe.options.rho * self.gradient_p(op, gradF)
        grad_pos = self.parameter_group.probe_positions.options.rho * self.gradient_pos(
            op, p, pos, gradF
        )

        if self.current_epoch == 0 or self.options.method == "GD":
            eta_o = -grad_o
            eta_op = -grad_op
            eta_p = -grad_p
            eta_pos = -grad_pos
        elif self.options.method == "CG":
            # intermediate variables for the CG
            eta_o = self.eta_o
            eta_op = self.eta_op[self.indices]
            eta_p = self.eta_p
            eta_pos = self.eta_pos[self.indices]

            beta = self.calc_beta_object_probe_positions(
                p, grad_p, eta_p, grad_pos, eta_pos, d, gradF, psi_far, op, grad_op, eta_op
            )
            eta_o = -grad_o + beta * eta_o
            eta_op = -grad_op + beta * eta_op
            eta_p = -grad_p + beta * eta_p
            eta_pos = -grad_pos + beta * eta_pos

        if self.options.method == "CG":
            self.eta_o = eta_o
            self.eta_op[self.indices] = eta_op
            self.eta_p = eta_p
            self.eta_pos[self.indices] = eta_pos

        alpha = self.calc_alpha_object_probe_positions(
            p, grad_o, grad_p, grad_pos, eta_o, eta_p, eta_pos, d, gradF, psi_far, op, eta_op
        )

        # normalize with the batch size
        alpha *= op.shape[0] / self.batch_size

        delta_o = alpha * eta_o
        delta_p = self.parameter_group.probe.options.rho * alpha * eta_p
        delta_pos = self.parameter_group.probe_positions.options.rho * alpha * eta_pos

        return (delta_o, delta_p, delta_pos)

    def gradientF(self, psi_far, d):
        """Gradient for the Guassian model"""

        td = d * (psi_far / (torch.abs(psi_far) + self.eps))
        td = psi_far - td
        # NOTE: scaling, needed to make fft adjoint for far field ptycho, should be removed for near field ptycho
        td *= psi_far.shape[-1] * psi_far.shape[-2]
        res = 2 * self.forward_model.free_space_propagator.propagate_backward(td)
        return res

    def hessianF(self, psi_far, psi_far1, psi_far2, data):
        """Hessian for the Guassian model"""

        l0 = psi_far / (torch.abs(psi_far) + self.eps)
        d0 = data / (torch.abs(psi_far) + self.eps)
        v1 = torch.sum((1 - d0) * reprod(psi_far1, psi_far2))
        v2 = torch.sum(d0 * reprod(l0, psi_far1) * reprod(l0, psi_far2))
        return 2 * (v1 + v2)

    def gradient_o(self, p, gradF):
        """Gradient with respect to the object"""

        tmp = torch.conj(p) * gradF

        obj = self.parameter_group.object
        o = obj.place_patches_on_empty_buffer(
            self.positions, 
            tmp[:, 0], 
            pad_for_shift=self.options.forward_model_options.pad_for_shift
        )
        patches = obj.extract_patches_function(
            o, self.positions + obj.pos_origin_coords, 
            self.parameter_group.probe.get_spatial_shape(), 
            pad=self.options.forward_model_options.pad_for_shift
        )
        patches = patches[:, None]
        return o, patches

    def gradient_p(self, op, gradF):
        """Gradient with respect to the probe"""

        return torch.sum(torch.conj(op) * gradF, axis=0)

    def gradient_pos(self, op, p, pos, gradF):
        """Gradient with respect to positions"""

        xi1 = torch.fft.fftfreq(op.shape[-1])
        xi2, xi1 = torch.meshgrid(xi1, xi1, indexing="xy")

        tmp = torch.fft.fft2(op)
        dt1 = 2 * torch.pi * 1j * torch.fft.ifft2(xi1 * tmp)
        dt2 = 2 * torch.pi * 1j * torch.fft.ifft2(xi2 * tmp)

        grad_pos = torch.zeros([len(pos), 2])
        grad_pos[:, 0] = redot(gradF, p * dt1, axis=(1, 2, 3))
        grad_pos[:, 1] = redot(gradF, p * dt2, axis=(1, 2, 3))

        return grad_pos

    def calc_alpha_object(self, p, do1, do2, d, psi_far, dop2):
        """Step length for the object update"""

        top = -redot(do1, do2)

        dm2 = p * dop2
        Ldm2 = self.forward_model.free_space_propagator.propagate_forward(dm2)
        bottom = self.hessianF(psi_far, Ldm2, Ldm2, d)
        return top / bottom

    def calc_alpha_object_probe(self, p, do1, dp1, do2, dp2, d, gradF, psi_far, op, dop2):
        """Step length for the object and probe update"""

        top = -redot(do1, do2) - redot(dp1, dp2)
        dp2 = self.parameter_group.probe.options.rho * dp2

        dm2 = dp2 * op + p * dop2
        d2m2 = 2 * dp2 * dop2
        Ldm2 = self.forward_model.free_space_propagator.propagate_forward(dm2)
        bottom = redot(gradF, d2m2) + self.hessianF(psi_far, Ldm2, Ldm2, d)
        return top / bottom

    def calc_alpha_object_probe_positions(
        self, p, do1, dp1, dpos1, do2, dp2, dpos2, d, gradF, psi_far, op, dop2
    ):
        """Step length for the object, probe and positions update"""

        top = -redot(do1, do2) - redot(dp1, dp2) - torch.sum(dpos1 * dpos2)

        dp2 = self.parameter_group.probe.options.rho * dp2
        dpos2 = self.parameter_group.probe_positions.options.rho * dpos2

        xi1 = torch.fft.fftfreq(p.shape[-1])
        [xi2, xi1] = torch.meshgrid(xi1, xi1, indexing="xy")

        dpos2 = dpos2[:, :, None, None, None]
        w1 = xi1 * dpos2[:, 0] + xi2 * dpos2[:, 1]
        w2 = (
            xi1**2 * dpos2[:, 0] ** 2
            + 2 * xi1 * xi2 * (dpos2[:, 0] * dpos2[:, 1])
            + xi2**2 * dpos2[:, 1] ** 2
        )

        tmp = torch.fft.fft2(dop2)
        dt2 = 2 * torch.pi * 1j * torch.fft.ifft2(w1 * tmp)

        tmp = torch.fft.fft2(op)
        dt = 2 * torch.pi * 1j * torch.fft.ifft2(w1 * tmp)
        d2t = -4 * torch.pi**2 * torch.fft.ifft2(w2 * tmp)

        dm2 = dp2 * op + p * (dop2 + dt)
        d2m2 = p * (2 * dt2 + d2t) + 2 * dp2 * dop2 + 2 * dp2 * dt

        Ldm2 = self.forward_model.free_space_propagator.propagate_forward(dm2)
        bottom = redot(gradF, d2m2) + self.hessianF(psi_far, Ldm2, Ldm2, d)
        return top / bottom

    def calc_beta_object(self, p, d, psi_far, dop1, dop2):
        """Step length for the CG direction with respect to the object"""

        dm1 = p * dop1
        dm2 = p * dop2

        Ldm1 = self.forward_model.free_space_propagator.propagate_forward(dm1)
        Ldm2 = self.forward_model.free_space_propagator.propagate_forward(dm2)

        top = self.hessianF(psi_far, Ldm1, Ldm2, d)
        bottom = self.hessianF(psi_far, Ldm2, Ldm2, d)
        return top / bottom

    def calc_beta_object_probe(self, p, dp1, dp2, d, gradF, psi_far, op, dop1, dop2):
        """Step length for the CG direction with respect to the object and probe"""

        dp1 = self.parameter_group.probe.options.rho * dp1
        dp2 = self.parameter_group.probe.options.rho * dp2

        dm1 = dp1 * op + p * dop1
        dm2 = dp2 * op + p * dop2

        d2m1 = dp1 * dop2 + dp2 * dop1
        d2m2 = 2 * dp2 * dop2

        Ldm1 = self.forward_model.free_space_propagator.propagate_forward(dm1)
        Ldm2 = self.forward_model.free_space_propagator.propagate_forward(dm2)

        top = redot(gradF, d2m1) + self.hessianF(psi_far, Ldm1, Ldm2, d)
        bottom = redot(gradF, d2m2) + self.hessianF(psi_far, Ldm2, Ldm2, d)
        return top / bottom

    def calc_beta_object_probe_positions(
        self, p, dp1, dp2, dpos1, dpos2, d, gradF, psi_far, op, dop1, dop2
    ):
        """Step length for the CG direction with respect to the object, probe and positions"""

        dp1 = self.parameter_group.probe.options.rho * dp1
        dp2 = self.parameter_group.probe.options.rho * dp2

        dpos1 = self.parameter_group.probe_positions.options.rho * dpos1
        dpos2 = self.parameter_group.probe_positions.options.rho * dpos2

        xi1 = torch.fft.fftfreq(p.shape[-1])
        [xi2, xi1] = torch.meshgrid(xi1, xi1, indexing="xy")

        dpos1 = dpos1[:, :, None, None, None]
        dpos2 = dpos2[:, :, None, None, None]

        w1 = xi1 * dpos1[:, 0] + xi2 * dpos1[:, 1]
        w2 = xi1 * dpos2[:, 0] + xi2 * dpos2[:, 1]
        w12 = (
            xi1**2 * dpos1[:, 0] * dpos2[:, 0]
            + xi1 * xi2 * (dpos1[:, 0] * dpos2[:, 1] + dpos1[:, 1] * dpos2[:, 0])
            + xi2**2 * dpos1[:, 1] * dpos2[:, 1]
        )
        w22 = (
            xi1**2 * dpos2[:, 0] ** 2
            + 2 * xi1 * xi2 * (dpos2[:, 0] * dpos2[:, 1])
            + xi2**2 * dpos2[:, 1] ** 2
        )

        tmp = torch.fft.fft2(dop1)
        dt12 = 2 * torch.pi * 1j * torch.fft.ifft2(w2 * tmp)

        tmp = torch.fft.fft2(dop2)
        dt21 = 2 * torch.pi * 1j * torch.fft.ifft2(w1 * tmp)
        dt22 = 2 * torch.pi * 1j * torch.fft.ifft2(w2 * tmp)

        tmp = torch.fft.fft2(op)
        dt1 = 2 * torch.pi * 1j * torch.fft.ifft2(w1 * tmp)
        dt2 = 2 * torch.pi * 1j * torch.fft.ifft2(w2 * tmp)

        d2t1 = -4 * torch.pi**2 * torch.fft.ifft2(w12 * tmp)
        d2t2 = -4 * torch.pi**2 * torch.fft.ifft2(w22 * tmp)

        dm1 = dp1 * op + p * (dop1 + dt1)
        dm2 = dp2 * op + p * (dop2 + dt2)

        d2m1 = p * (dt12 + dt21 + d2t1) + dp1 * (dop2 + dt2) + dp2 * (dop1 + dt1)
        d2m2 = p * (2 * dt22 + d2t2) + 2 * dp2 * dop2 + 2 * dp2 * dt2

        Ldm1 = self.forward_model.free_space_propagator.propagate_forward(dm1)
        Ldm2 = self.forward_model.free_space_propagator.propagate_forward(dm2)

        top = redot(gradF, d2m1) + self.hessianF(psi_far, Ldm1, Ldm2, d)
        bottom = redot(gradF, d2m2) + self.hessianF(psi_far, Ldm2, Ldm2, d)
        return top / bottom
